/*JS*********************************************************************
*
*    Program : JSDLLIST
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Doubly linked lists.
*
*              Copyright (C) 2006-2013 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/
#ifndef JSDLLIST_H /*   File must not be included more than once!   */
#define JSDLLIST_H 1

#include <stddef.h> /*  "offsetof" macro and "NULL"  */

/* ***  A doubly linked list node  *** */
typedef struct _jsdllnode_t _jsdllnode_t;

struct _jsdllnode_t {
  /*  The sequence of these fields must not be changed!  */
  _jsdllnode_t *jdlln_Next;
  _jsdllnode_t *jdlln_Prev;

};

/* ***  A doubly linked list header node  *** */
typedef struct {
  /*
   *  The first and second entry comprise Next + Prev field of the
   *   pseudo 'header' node.
   *   The second and third one comprise the Next + Prev field of
   *   the pseudo 'tail' node.
   *   The second field is therefore always NULL.
   */

  /*  The sequence of these fields must not be changed!  */
  _jsdllnode_t *jdllh_HeadNode;
  _jsdllnode_t *jdllh_Dummy;
  _jsdllnode_t *jdllh_TailNode;

} _jsdll_t;

/* ***  Initialize a header node  *** */
#define _jsdllinit(hptr)						\
  ((hptr)->jdllh_HeadNode = (_jsdllnode_t *)&((hptr)->jdllh_Dummy),	\
   (hptr)->jdllh_Dummy    = NULL,					\
   (hptr)->jdllh_TailNode = (_jsdllnode_t *)&((hptr)->jdllh_HeadNode))

#define _jsdllisempty(hptr)						\
  ((hptr)->jdllh_HeadNode == (_jsdllnode_t *)&((hptr)->jdllh_Dummy) &&	\
   (hptr)->jdllh_TailNode == (_jsdllnode_t *)&((hptr)->jdllh_HeadNode))

/* ***  Access a list node  *** */
#define _jsdllgetprev(nptr)        ((nptr)->jdlln_Prev)
#define _jsdllgetnext(nptr)        ((nptr)->jdlln_Next)

/*  These macros give access to internal fields  */
#define _jsdllsetprev(nptr, mptr)  ((nptr)->jdlln_Prev = (mptr))
#define _jsdllsetnext(nptr, mptr)  ((nptr)->jdlln_Next = (mptr))


/* ***  From a header to the start or end of the list   *** */
#define _jsdllgetfirstnode(hptr)   ((hptr)->jdllh_HeadNode)
#define _jsdllgetlastnode(hptr)    ((hptr)->jdllh_TailNode)

/*  These macros give access to internal fields  */
#define _jsdllsetfirstnode(hptr, nptr)					\
  ((hptr)->jdllh_HeadNode = (nptr),					\
   (nptr)->jdlln_Prev = (_jsdllnode_t *)&((hptr)->jdllh_HeadNode))

#define _jsdllsetlastnode(hptr, nptr)				\
  ((hptr)->jdllh_TailNode = (nptr),				\
   (nptr)->jdlln_Next = (_jsdllnode_t *)&((hptr)->jdllh_Dummy))

/* ***  If we reach the head or tail of a list, how to move to the list head  *** */
#define _jsdllhead2dll(nptr)       ((_jsdll_t *)&((nptr)->jdlln_Next))
#define _jsdlltail2dll(nptr)       ((_jsdll_t *)(&((nptr)->jdlln_Next) - 1))

/* ***  Add a node at the tail  *** */
#define _jsdlladdtail(hptr, nptr)					\
  _jsdlladdbefore((_jsdllnode_t *)&((hptr)->jdllh_Dummy), nptr)

/* ***  Add a node at the head  *** */
#define _jsdlladdhead(hptr, nptr)					\
  _jsdlladdafter((_jsdllnode_t *)&((hptr)->jdllh_HeadNode), nptr)

/*  Add a node before/after another node  */
#define _jsdlladdbefore(mptr, nptr)			      \
  ((nptr)->jdlln_Next             = (mptr),		      \
   (nptr)->jdlln_Prev             = (mptr)->jdlln_Prev,	      \
   (mptr)->jdlln_Prev->jdlln_Next = (nptr),		      \
   (mptr)->jdlln_Prev             = (nptr))
#define _jsdlladdafter(mptr, nptr)			      \
  ((nptr)->jdlln_Prev             = (mptr),		      \
   (nptr)->jdlln_Next             = (mptr)->jdlln_Next,	      \
   (mptr)->jdlln_Next->jdlln_Prev = (nptr),		      \
   (mptr)->jdlln_Next             = (nptr))

/*  Add a whole list at the head/tail of another and empty the list afterwards  */
#define _jsdlllistaddtail(hptr, lptr)					\
  _jsdlllistaddbefore((_jsdllnode_t *)&((hptr)->jdllh_Dummy), lptr)

#define _jsdlllistaddhead(hptr, lptr)					\
  _jsdlllistaddafter((_jsdllnode_t *)&((hptr)->jdllh_Dummy), lptr)

/*  Add a whole list before/after a node and empty the list afterwards  */
#define _jsdlllistaddbefore(mptr, lptr)					\
	((lptr)->jdllh_HeadNode->jdlln_Prev = (mptr)->jdlln_Prev,	\
	 (lptr)->jdllh_TailNode->jdlln_Next = (mptr),			\
	 (mptr)->jdlln_Prev->jdlln_Next     = (lptr)->jdllh_HeadNode,	\
	 (mptr)->jdlln_Prev                 = (lptr)->jdllh_TailNode,	\
	 (lptr)->jdllh_HeadNode = (_jsdllnode_t *)&((lptr)->jdllh_Dummy), \
	 (lptr)->jdllh_TailNode = (_jsdllnode_t *)&((lptr)->jdllh_HeadNode))
#define _jsdlllistaddafter(mptr, lptr)					\
  ((lptr)->jdllh_HeadNode->jdlln_Prev = (mptr),				\
   (lptr)->jdllh_TailNode->jdlln_Next = (mptr)->jdlln_Next,		\
   (mptr)->jdlln_Next->jdlln_Prev     = (lptr)->jdllh_TailNode,		\
   (mptr)->jdlln_Next                 = (lptr)->jdllh_HeadNode,		\
   (lptr)->jdllh_HeadNode = (_jsdllnode_t *)&((lptr)->jdllh_Dummy),	\
   (lptr)->jdllh_TailNode = (_jsdllnode_t *)&((lptr)->jdllh_HeadNode))

/*  Remove a node from a list  */
#define _jsdllremove(nptr)				      \
  ((nptr)->jdlln_Prev->jdlln_Next = (nptr)->jdlln_Next,	      \
   (nptr)->jdlln_Next->jdlln_Prev = (nptr)->jdlln_Prev)

#define _jsdllishead(nptr)         ((nptr)->jdlln_Prev == NULL)
#define _jsdllistail(nptr)         ((nptr)->jdlln_Next == NULL)

/* ***  Special node that is not in a list  *** */ 
#define _jsdllcleannode(nptr)				\
  ((nptr)->jdlln_Prev = (nptr)->jdlln_Next = NULL)

/*  How to check for a special node  */
#define _jsdllnodeinlist(nptr)					\
  ((nptr)->jdlln_Prev != NULL || (nptr)->jdlln_Next != NULL)

/* ***  How to convert from the linked list member to the actual user type and back  *** */
#define _jsdllnode2utype(nptr, utype, member)		\
    ((utype *)((char *)(nptr) - offsetof(utype, member)))
#define _jsdllutype2node(nptr, utype, member)				\
    ((_jsdllnode_t *)((char *)(nptr) + offsetof(utype, member)))

/* ***  How to adapt headers and nodes that are moved in memory  *** */
/*  This is rather complex since it must also work for empty lists!  */
#define _jsdllmoveheader(hptr)		       \
  (((unsigned char *)(hptr)->jdllh_HeadNode -  \
    offsetof(_jsdll_t, jdllh_Dummy)) ==	       \
   ((unsigned char *)(hptr)->jdllh_TailNode -  \
    offsetof(_jsdll_t, jdllh_HeadNode)) ?      \
   _jsdllinit(hptr) :			       \
   ((hptr)->jdllh_HeadNode->jdlln_Prev =       \
    (_jsdllnode_t *)&((hptr)->jdllh_HeadNode), \
    (hptr)->jdllh_TailNode->jdlln_Next =       \
    (_jsdllnode_t *)&((hptr)->jdllh_Dummy)))

#define _jsdllmovenode(nptr)			  \
  ((nptr)->jdlln_Prev->jdlln_Next = (nptr),	  \
   (nptr)->jdlln_Next->jdlln_Prev = (nptr))

/* ***  Loop over lists forward and backward  *** */
#define _jsdllloop(hptr, nptr, utype, member)				\
  for(nptr = _jsdllnode2utype(_jsdllgetfirstnode(hptr), utype, member) ; \
      !_jsdllistail(_jsdllutype2node(nptr, utype, member)) ;		\
      nptr = _jsdllnode2utype(_jsdllgetnext(_jsdllutype2node(nptr,	\
							     utype, member)), \
			      utype, member))

#define _jsdlllooprev(hptr, nptr, utype, member)			\
	for(nptr = _jsdllnode2utype(_jsdllgetlastnode(hptr), utype, member) ; \
	    !_jsdllishead(_jsdllutype2node(nptr, utype, member)) ;	\
	    nptr = _jsdllnode2utype(_jsdllgetprev(_jsdllutype2node(nptr, \
								   utype, member)), \
				    utype, member))

#endif
