/*JS*********************************************************************
*
*    Program : JSSUBS.H
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Include file for library 'libjssubs.a' and all programs
*	       that use this library.
*
*              Copyright (C) 1994-2015 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/
#ifndef JSSUBS_H /*   File must not be included more than once!   */
#define JSSUBS_H 1

/*********     GENERAL DEFINES					*********/
#ifndef FALSE
# define FALSE	0
#endif
#ifndef TRUE
# define TRUE	1
#endif

/* ****    Some useful macros (be aware of side effects because   **** */
/* ****     they treat arguments more than once).		  **** */
#ifndef MAX
# define MAX(a,b)  ((a) > (b) ? (a) :  (b))
#endif
#ifndef MIN
# define MIN(a,b)  ((b) > (a) ? (a) :  (b))
#endif
#ifndef ABS
# define ABS(x)    ((x) >= 0  ? (x) : -(x))
#endif

/* ********************************************************************** */
#ifdef __GNUC__
# ifndef likely
#  define likely(x)     (__builtin_expect(!!(x), 1))
# endif
# ifndef unlikely
#  define unlikely(x)   (__builtin_expect(!!(x), 0))
# endif
#else
# ifndef likely
#  define likely(x)     (x)
# endif
# ifndef unlikely
#  define unlikely(x)   (x)
# endif
#endif

/* ********************************************************************** */
/*  This is used to reallocate dynamic arrays in a smart way:
 *   We increase the size by 25%
 */
#define REALLOC_RESIZE(maxvar, incr)  ((maxvar) += MAX((maxvar) / 4, incr))

/* ********************************************************************** */
#ifndef EGOTO
# ifdef DEBUG_EGOTO_VERBOSE0
#  define EGOTO(label)							\
    do {								\
	fprintf(stderr, "ERROR l. %d '%s'\n", __LINE__, __FILE__);	\
	goto label;							\
    } while(0)
# elif defined(DEBUG_EGOTO_VERBOSE)
#  define EGOTO(label)							\
    do {								\
	sprintf(JSErrorStack[JSErrorStackLevel], "l. %d '%s'", __LINE__, __FILE__); \
	if(++JSErrorStackLevel >= sizeof(JSErrorStack) / sizeof(JSErrorStack[0])) \
	    JSErrorStackLevel = 0;					\
	goto label;							\
    } while(0)
# elif defined(DEBUG_EGOTO_ABORT)
#  define EGOTO(label)    abort()
# else
#  define EGOTO(label)    goto label
# endif
#endif

/*  This will initialize the error call stack (not yet)  */
#define EGOTO0(label)     EGOTO(label)

/* ********************************************************************** */
/*  Mark error if's. For GCC, we help branch prediction
 *   We could also redefine this to force errors
 */
#ifndef EIF
# define EIF(expr)         if(unlikely(expr))
#endif

/* ********************************************************************** */
/* ****   Compile time helper   ***** */
/*  Test a condition at compile time instead of runtime. Idea taken from
 *   the Linux kernel
 */
#define CASSERT(condition)        ((void)sizeof(int [1 - 2 * !(condition)]))

/* ********************************************************************** */
#define USE(param)        (void)param

/* ********************************************************************** */
/* ****   I'd like to type PI if I want PI !   ***** */
#ifndef PI
# define PI         3.1415926535897932384626433832795029
#endif

#ifndef LN2
# define LN2        0.6931471805599453094172321214581766
#endif

/* ********************************************************************** */
/* ****   Interval intersection   ***** */
/*  Check if two intervals intersect. It is easy if you have the
 *   right perspective on it: They *don't* intersect if the first
 *   intervals end is left of the second intervals start or vice
 *   versa. Reversing the condition gives the macros below.
 */

/*  Intersection including boundaries  */
#define INTERVAL_INTERSECT(astart, aend, bstart, bend) \
		((aend) >= (bstart) && (bend) >= (astart))

/*  Intersection excluding boundaries  */
#define OPEN_INTERVAL_INTERSECT(astart, aend, bstart, bend) \
		((aend) > (bstart) && (bend) > (astart))

/* ********************************************************************** */
/*  Calculate required padding for alignment  */
#define JS_ALIGN_OFFSET(offset, align)		\
    ((offset) % (align) == 0 ? (offset) :	\
     (offset) + (align) - ((offset) % (align)))

/* *******  General definitions  *********** */
/*  Some systems define FILENAME_MAX very small (14 for example)  */
#include <stdio.h>

#ifndef JS_FILENAME_MAX
# if !defined(FILENAME_MAX) || (FILENAME_MAX < 256 && !defined(__MSDOS__))
#  define JS_FILENAME_MAX  256
# else
#  define JS_FILENAME_MAX  FILENAME_MAX
# endif
#endif

/* ******  Common environment variable for options and path ***** */
#define JSENVOPT     "JSOPTIONS"
#define JSENVPATH    "JSPATH"

/* ***********	 Extracted definitions of library code  ********* */
/* ----------- FILE: bigsort.c            ------------ */
#include <stdio.h>   /*  For "FILE" definition    */
#include <jsstdint.h>

typedef jsuint64_t bigsortcount_t;

typedef unsigned char bigsortdata_t;

typedef struct
{
    bigsortdata_t *BSIE_Data;
    unsigned int   BSIE_Length;
}
bigsortentry_t;

/*  Structure to access a sorted stream  */
typedef struct {
    bigsortdata_t *BSG_Previous;
    unsigned int   BSG_PrevLen;

    int (*BSG_Compare)(void *uptr,
                       const void *a, unsigned int la,
                       const void *b, unsigned int lb);
    void *BSG_User;

} bigsortgroup_t;

typedef struct {
    unsigned int   BSH_Type;

#define BIGSORT_TYPE_TCOLLECTING   0 /*  still builing up  */
#define BIGSORT_TYPE_TMEMLIST      1 /*  memory list ("BSH_Buff")  */
#define BIGSORT_TYPE_TFILELIST     2 /*  file access ("BSH_File")  */
#define BIGSORT_TYPE_TMASK         3

#define BIGSORT_TYPE_FKEEPSTRUCT   (1 << 3) /*  do not free structure itself - it was passed  */
#define BIGSORT_TYPE_FVARLENGTH    (1 << 4) /*  variable length stream  */
#define BIGSORT_TYPE_FVARLENGTH2   (1 << 5) /*  also uses variable length integers  */
#define BIGSORT_TYPE_FGROUPING     (1 << 6) /*  BSH_Group is for grouping, not for distinct  */
#define BIGSORT_TYPE_FFILEPOSOK    (1 << 7) /*  BS_Current matches current file position  */
#define BIGSORT_TYPE_FINVERTED     (1 << 8)

    unsigned int   BSH_Len;
    bigsortcount_t BSH_Count, BSH_Curr;
    union {
        void           *bsh_Work; /*  during built-up, points to "bigsort_t"  */
        bigsortentry_t *bsh_Buff; /*  for memory list  */
        FILE           *bsh_File; /*  for file access  */
    } c;

    bigsortgroup_t *BSH_Group;

    unsigned int    BSH_NTempFiles, BSH_MaxLevel;
    bigsortcount_t  BSH_TotalWritten, BSH_TotalRead, BSH_MaxTempSize;
    bigsortcount_t  BSH_TotalRuns;

} bigsorthandle_t;

#define BSH_Work  c.bsh_Work
#define BSH_File  c.bsh_File
#define BSH_Buff  c.bsh_Buff

/* **  For variable mode from bigsortopen  ** */
#define BIGSORT_FDISTINCT       (1 << 0) /* remove duplicate entries  */
#define BIGSORT_FDISTINCT2      (1 << 1) /* remove duplicate entries and provide direct access  */
#define BIGSORT_FGROUP          (1 << 2) /* do grouping  */
#define BIGSORT_FVARLEN         (1 << 3) /* input data has variable length,
                                          *  length argument only gives maximum
                                          *  (for buffer allocation)
                                          */
#define BIGSORT_FVARLEN2        (1 << 4) /* same as BIGSORT_FVARLEN2, but uses
					  *  variable length integers
					  */
#define BIGSORT_FINVERTED       (1 << 5)

/*  Obtain length of sorted stream  */
#define bigsortlength(bsh)      ((bsh)->BSH_Count)

/*  Seek or rewind a stream and get current position  */
#define bigsortseek(bsh, pos)    ((bsh)->BSH_Curr = \
                                  (pos) != (unsigned int)-1 ? (pos) : 0)
#define bigsortrewind(bsh)      bigsortseek(bsh, 0)
#define bigsortcurrpos(bsh)     ((bsh)->BSH_Curr)

/*  Get next set in sequence (sequential access)  */
#define bigsortnext(bsh, buff, lbuff)           \
    bigsortget(bsh, -1, buff, lbuff)

/*  Same as bigsortnext, but do not advance internal pointer  */
#define bigsortpeek(bsh, buff, lbuff)           \
    bigsortget(bsh, -2, buff, lbuff)

/*  Portable way to check if this handle does grouping  */
#define bigsortisgroup(bsh)                     \
    ((bsh)->BSH_Group != NULL)

/* ----------- FILE: jserror.c            ------------ */
#include <errno.h>

#define JSCLEARSYSERROR()       (errno = 0)

#define JSERROR_MAXLENGTH       1000

#ifndef _ERRMODE_STRING

# ifdef DEBUG_EGOTO_VERBOSE
#  define JSCLEARERROR()     (JSErrorStackLevel = 0,	\
			      JSErrNo = 0,		\
			      JSErrString[0] = '\0')
# else
#  define JSCLEARERROR()     (JSErrNo = 0,		\
			      JSErrString[0] = '\0')
# endif

# define JSSETERROR(code, off)						\
  (JSCLEARSYSERROR(),							\
   JSErrNo = (code) + (off), JSErrLine = 0,				\
   JSErrString[0] = '\0')

# define JSSETERROR2X(code, off, string)				\
  (JSCLEARSYSERROR(),							\
   JSErrNo = (code) + (off), JSErrLine = 0,				\
   strncpy(JSErrString, string, sizeof(JSErrString) - 1))
#else
# ifdef DEBUG_EGOTO_VERBOSE
#  define JSCLEARERROR()						\
  (JSErrorStackLevel = 0,						\
   JSErrNo = JSErrLine = 0, JSErrLabel = NULL,				\
   JSErrString[0] = '\0')
# else
#  define JSCLEARERROR()					\
  (JSErrNo = JSErrLine = 0, JSErrLabel = NULL,			\
   JSErrString[0] = '\0')
# endif

# define JSSETERROR(code, off)						\
  (JSCLEARSYSERROR(),							\
   JSErrLabel = #code, JSErrNo = 1 + (off),				\
   JSErrLine = 0, JSErrString[0] = '\0')

# define JSSETERROR2X(code, off, string)				\
  (JSCLEARSYSERROR(),							\
   JSErrLabel = #code, JSErrNo = 1 + (off),				\
   strncpy(JSErrString, string, sizeof(JSErrString) - 1))
#endif

# define JSSETERRSTRING(string)					\
  strncpy(JSErrString, string, sizeof(JSErrString) - 1)

# define JSSETSYSERRSTRING(string)			\
  (JSCLEARERROR(), JSSETERRSTRING(string))

#define JSSETERROR2(code, off, string)				\
  JSSETERROR2X(code, off, (string) ? (string) : "<NULL>")

#define JSSETERRFILE(file, line)		\
  JSSETERRFILE2(file, line, 1)

#define JSSETERRFILE2(file, line, column)				\
  (strncpy(JSErrFile, (file) ? (file) : "<NULL>", sizeof(JSErrFile) - 1), \
   JSErrLine = (line), JSErrColumn = (column))

/*  These are internal error codes for debugging  */
#define JSMINIERROR             10000
#define JSSETIERROR(code, off)   JSSETERROR(JSMINIERROR + (code), off)
#define JSSETIERROR2(code, off, s) JSSETERROR2(JSMINIERROR + (code), off, s)

/* ----------- FILE: ErrCodes.c           ------------ */
#define _ERR_BIGSORTOPEN            1
#define _ERR_BIGSORTPUT             2
#define _ERR_BIGSORTGET             3

#define _ERR_MAXERROR               6

/* ----------- FILE: TraceCodes.c         ------------ */
#ifndef CONFIG_JSTRACE_NOTRACES
# define TRACETYPE_BIGSORT           0
# define TRACETYPE_MAIN              1
#endif


/* *******   Prototypes of library functions  ***** */
/*  For C++ compatibility   */
# ifdef __cplusplus
#  define Prototype extern "C"
# else
#  define Prototype extern
# endif
/* ***** File: bigsort.c             ***** */
Prototype bigsorthandle_t *bigsortopen(bigsorthandle_t *bsh, unsigned int len,
                                       bigsortcount_t nExpect,
                                       int (*doget)(void *uptr,
                                                    void *buff, unsigned int lbuff,
                                                    bigsortcount_t nr),
                                       int (*compare)(void *uptr,
                                                      const void *a, unsigned int la,
                                                      const void *b, unsigned int lb),
                                       void *uPtr, int mode);
Prototype int            bigsortclose(bigsorthandle_t *bsh);
Prototype bigsortcount_t bigsortput(bigsorthandle_t *bsh, const void *buff,
                                    unsigned int lbuff);
Prototype int            bigsortget(bigsorthandle_t *bsh, bigsortcount_t nr,
                                    void *buff, unsigned int lbuff);
Prototype bigsortcount_t bigsortfind(bigsorthandle_t *bsh, const void *cont, unsigned int lcont,
                                     int (*compare)(void *uptr, const void *a, unsigned int la,
                                                    const void *b, unsigned int lb), void *uPtr,
                                     void *dest);
Prototype jsuint64_t     BSortMaxMem;
Prototype unsigned int   BSortNSeqs, BSSortFlags;
Prototype const char    *BSortTempDir;
Prototype FILE        *(*BSortGetTempFile)(void *, unsigned int nr);
Prototype void          *BSortGetTempFilePtr;

/* ***** File: debug.c               ***** */
Prototype long           Debug;

/* ***** File: jserror.c             ***** */
Prototype int            JSErrNo;
Prototype char           JSErrString[JSERROR_MAXLENGTH + 1];
Prototype const char    *JSErrLabel;
Prototype int            JSErrLine, JSErrColumn;
Prototype char           JSErrFile[JS_FILENAME_MAX + 1];
Prototype int            JSErrorStackLevel;
Prototype char           JSErrorStack[100][JSERROR_MAXLENGTH];

/* ***** File: jsgetopt.c            ***** */
Prototype int		 jsGetOpt(int *pargc, void *argv, const char *optstring,
				  const char **poptarg);
Prototype int            jsFirstArg(char *argv[], const char *optstring);
Prototype int		 JSOptMode, JSOptIndex;

/* ***** File: jsperror.c            ***** */
Prototype void           jsperror(const char *progname);

/* ***** File: jsstrerr.c            ***** */
Prototype const char    *jsStrError(int err);
Prototype int            JSErrorMode;

/* ***** File: jstrace.c             ***** */
Prototype void           jstrace(int type, int level,
				 const char *format, ...);
Prototype FILE          *jstracegetfp(int type, int level);
Prototype int            jstracelevel(int type);
Prototype int            jstracemask(int type);
Prototype int            jstraceset(const char *type, int level);
Prototype int            jstraceseti(int ttype, int level);
Prototype int            jstracesets(const char *string);
Prototype char           JSTraceDirectory[JS_FILENAME_MAX];
Prototype int            JSTraceFlags;

/* ***** File: strcomp.c             ***** */
Prototype const char    *jstrstartswith(const char *input, const char *string);
Prototype const char    *jstristartswith(const char *input, const char *string);
Prototype int            jstrendswith(const char *input, const char *string);

/* ***** File: ErrCodes.c            ***** */
Prototype const char * const JSErrors[];
Prototype const int          JSErrorsI[];

/* ***** File: TraceCodes.c          ***** */
Prototype const char * const JSTraceTypes[];
Prototype int                JSTraceLevels[];
Prototype int                JSNTraceTypes;


# undef Prototype   /* I like to be clean  */

#endif /*  "#ifdef JSSUBS_H"  */
