/*JS*********************************************************************
*
*    Program : BIGSORT
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Sort huge amounts of data by doing a multi-phase
*              sorting on temporary files.
*
*   For a description of the algorithm see Section 2.4.4 in
*   Niklaus Wirth: "Algorithmen und Datenstrukturen mit Modula-2"
*   The algorithm was enhanced after an idea taken from Sedgewick
*   and now uses a heap to keep track of the minimum element while
*   merging sequences.
*
*   This implementation automatically recognizes if the input data fits
*   completely into memory and then does a in-place sort via quicksort.
*
*              Copyright (C) 1994-2016 Joerg Schoen
*
* -----------------------------------------------------------------------
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDE                                         *********/
#include <jsalloca.h>

#include <assert.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <jsstdint.h>

#include <jsconfig.h>

#include <jssubs.h>

#ifndef CONFIG_NO_POSIX
#include <jsllio.h>
#endif

#ifdef __linux__
# define fseek              fseeko
#endif

/*********     DEFINES                                          *********/
#if 1
#define CONFIG_BIGSORT_USEVARINT /*  may use variable length integer to store length prefix  */
#endif

#if 1 /*  only used if BSortFlags bit 0 is set!  */
#define CONFIG_BIGSORT_USETRUNCATE /*  truncate output file before reuse  */
#endif

#define CONFIG_BIGSORT_USEGENSORT   /*  it is usually faster  */

#ifndef BSH_File
/*  START-DEFINITIONS  */
#include <stdio.h>   /*  For "FILE" definition    */
#include <jsstdint.h>

typedef jsuint64_t bigsortcount_t;

typedef unsigned char bigsortdata_t;

typedef struct
{
    bigsortdata_t *BSIE_Data;
    unsigned int   BSIE_Length;
}
bigsortentry_t;

/*  Structure to access a sorted stream  */
typedef struct {
    bigsortdata_t *BSG_Previous;
    unsigned int   BSG_PrevLen;

    int (*BSG_Compare)(void *uptr,
                       const void *a, unsigned int la,
                       const void *b, unsigned int lb);
    void *BSG_User;

} bigsortgroup_t;

typedef struct {
    unsigned int   BSH_Type;

#define BIGSORT_TYPE_TCOLLECTING   0 /*  still builing up  */
#define BIGSORT_TYPE_TMEMLIST      1 /*  memory list ("BSH_Buff")  */
#define BIGSORT_TYPE_TFILELIST     2 /*  file access ("BSH_File")  */
#define BIGSORT_TYPE_TMASK         3

#define BIGSORT_TYPE_FKEEPSTRUCT   (1 << 3) /*  do not free structure itself - it was passed  */
#define BIGSORT_TYPE_FVARLENGTH    (1 << 4) /*  variable length stream  */
#define BIGSORT_TYPE_FVARLENGTH2   (1 << 5) /*  also uses variable length integers  */
#define BIGSORT_TYPE_FGROUPING     (1 << 6) /*  BSH_Group is for grouping, not for distinct  */
#define BIGSORT_TYPE_FFILEPOSOK    (1 << 7) /*  BS_Current matches current file position  */
#define BIGSORT_TYPE_FINVERTED     (1 << 8)

    unsigned int   BSH_Len;
    bigsortcount_t BSH_Count, BSH_Curr;
    union {
        void           *bsh_Work; /*  during built-up, points to "bigsort_t"  */
        bigsortentry_t *bsh_Buff; /*  for memory list  */
        FILE           *bsh_File; /*  for file access  */
    } c;

    bigsortgroup_t *BSH_Group;

    unsigned int    BSH_NTempFiles, BSH_MaxLevel;
    bigsortcount_t  BSH_TotalWritten, BSH_TotalRead, BSH_MaxTempSize;
    bigsortcount_t  BSH_TotalRuns;

} bigsorthandle_t;

#define BSH_Work  c.bsh_Work
#define BSH_File  c.bsh_File
#define BSH_Buff  c.bsh_Buff

/* **  For variable mode from bigsortopen  ** */
#define BIGSORT_FDISTINCT       (1 << 0) /* remove duplicate entries  */
#define BIGSORT_FDISTINCT2      (1 << 1) /* remove duplicate entries and provide direct access  */
#define BIGSORT_FGROUP          (1 << 2) /* do grouping  */
#define BIGSORT_FVARLEN         (1 << 3) /* input data has variable length,
                                          *  length argument only gives maximum
                                          *  (for buffer allocation)
                                          */
#define BIGSORT_FVARLEN2        (1 << 4) /* same as BIGSORT_FVARLEN2, but uses
					  *  variable length integers
					  */
#define BIGSORT_FINVERTED       (1 << 5)

/*  Obtain length of sorted stream  */
#define bigsortlength(bsh)      ((bsh)->BSH_Count)

/*  Seek or rewind a stream and get current position  */
#define bigsortseek(bsh, pos)    ((bsh)->BSH_Curr = \
                                  (pos) != (unsigned int)-1 ? (pos) : 0)
#define bigsortrewind(bsh)      bigsortseek(bsh, 0)
#define bigsortcurrpos(bsh)     ((bsh)->BSH_Curr)

/*  Get next set in sequence (sequential access)  */
#define bigsortnext(bsh, buff, lbuff)           \
    bigsortget(bsh, -1, buff, lbuff)

/*  Same as bigsortnext, but do not advance internal pointer  */
#define bigsortpeek(bsh, buff, lbuff)           \
    bigsortget(bsh, -2, buff, lbuff)

/*  Portable way to check if this handle does grouping  */
#define bigsortisgroup(bsh)                     \
    ((bsh)->BSH_Group != NULL)
/*  END-DEFINITIONS  */
#endif

/* ************************************************************************ */
/* **********************    Internal defines     ************************* */
/* ************************************************************************ */
/*  Maximum allocated memory  */
#ifndef BIGSORT_MAXMEM
/*  Higher numbers are not necessarily better, so I stay there  */
# define BIGSORT_MAXMEM   (8 * 1024 * 1024) /*  8 MBytes  */
#endif

typedef unsigned int bigsortvarlen_t;

/*  Internal structures  */
typedef struct {
    FILE          *SEQ_File;
    bigsortcount_t SEQ_FileSize;
    bigsortdata_t *SEQ_Buffer;
    unsigned int   SEQ_BuffLen;

    /*  Number of elements in this sequence */
    bigsortcount_t SEQ_Count;
    /*  Number of runs in this sequence (actual + pseudo)  */
    bigsortcount_t SEQ_NRuns;
    /*  Number of pseudo runs contained in 'SEQ_NRuns'  */
    bigsortcount_t SEQ_DiffRuns;

} bigsortseq_t;

typedef struct {
#define BIGSORT_STATUS_DISTRUNSMEM    0
#define BIGSORT_STATUS_DISTRUNSEXT    1
    unsigned int     BSI_Status;
    unsigned int     BSI_Mode;

    bigsorthandle_t *BSI_Handle;

    bigsortcount_t   BSI_S, BSI_R;

    unsigned int     BSI_Len;
    void            *BSI_UPtr;
    int            (*BSI_Compare)(void *uptr,
				  const void *a, unsigned int la,
				  const void *b, unsigned int lb);

    bigsortcount_t   BSI_Count, BSI_NArray;
    bigsortseq_t    *BSI_Seqs;
    unsigned int     BSI_NSeqs;

    unsigned int     BSI_Level;
    unsigned int     BSI_Curr;  /*  current output sequence  */

    bigsortentry_t  *BSI_Array;

} bigsort_t;

/*  assuming most strict alignment is for type "double"  */
#define JSMAXALIGN         sizeof(double)

/*  TRACE-DEFINITION symbol TRACETYPE_BIGSORT name BIGSORT */

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype bigsorthandle_t *bigsortopen(bigsorthandle_t *bsh, unsigned int len,
                                       bigsortcount_t nExpect,
                                       int (*doget)(void *uptr,
                                                    void *buff, unsigned int lbuff,
                                                    bigsortcount_t nr),
                                       int (*compare)(void *uptr,
                                                      const void *a, unsigned int la,
                                                      const void *b, unsigned int lb),
                                       void *uPtr, int mode);
static bigsorthandle_t  *bigsortopen0(bigsorthandle_t *bsh, unsigned int len,
                                      bigsortcount_t nExpect,
                                      int (*compare)(void *uptr,
                                                     const void *a, unsigned int la,
                                                     const void *b, unsigned int lb),
                                      void *uPtr, int mode);
Prototype int            bigsortclose(bigsorthandle_t *bsh);

static int               optimizeMemList(bigsorthandle_t *bsh, bigsortentry_t *array,
                                         bigsortcount_t countold,
                                         bigsortcount_t countnew);
Prototype bigsortcount_t bigsortput(bigsorthandle_t *bsh, const void *buff,
                                    unsigned int lbuff);
static void              distElementMem(bigsort_t *pbs, const void *buff,
                                        unsigned int lbuff);

static int               setupExt(bigsort_t *pbs);
static int               distElementExt(bigsort_t *pbs, const void *buff,
                                        unsigned int lbuff);

static int               finalizeExt(bigsort_t * pbs);

static bigsortcount_t    doQuickSort(bigsort_t *pbs, int mode);

static void              bsSift(bigsort_t *pbs, bigsortcount_t l, bigsortcount_t r);
static void              bsSiftInd(bigsort_t *pbs, bigsortcount_t l, bigsortcount_t r,
                                   unsigned int *tr);

#ifdef CONFIG_BIGSORT_USEGENSORT
static void              sortSpecialVar(bigsortentry_t *array, bigsortcount_t n,
                                        int (*compare)(void *uptr, const void *a, unsigned int la,
                                                       const void *b, unsigned int lb),
                                        void *uPtr);
static void              sortSpecialVarInv(bigsortentry_t *array, bigsortcount_t n,
                                           int (*compare)(void *uptr, const void *a, unsigned int la,
                                                          const void *b, unsigned int lb),
                                           void *uPtr);
#else
static void             *CompareUPtr;
static int             (*CompareFct)(void *uptr, const void *a, unsigned int la,
                                     const void *b, unsigned int lb);
static unsigned int      CompareLen;

static int               sortCompare(const void *a, const void *b);
static int               sortCompareInv(const void *a, const void *b);
#endif

static int               storeContent(bigsorthandle_t *bsh, FILE * fp, bigsortcount_t *psize,
                                      const bigsortdata_t *data, bigsortvarlen_t dlen);
static int               readContent(bigsorthandle_t *bsh, FILE * fp, bigsortdata_t *data,
                                     bigsortvarlen_t *pdlen);

static int               selectRun(bigsort_t *pbs);

static int               mergeRuns(bigsort_t *pbs);

static int               removeduplicates(bigsort_t *pbs);

static FILE             *getTempFile0(unsigned int nr);

Prototype int            bigsortget(bigsorthandle_t *bsh, bigsortcount_t nr,
                                    void *buff, unsigned int lbuff);
Prototype bigsortcount_t bigsortfind(bigsorthandle_t *bsh, const void *cont, unsigned int lcont,
                                     int (*compare)(void *uptr, const void *a, unsigned int la,
                                                    const void *b, unsigned int lb), void *uPtr,
                                     void *dest);

Prototype jsuint64_t     BSortMaxMem;
Prototype unsigned int   BSortNSeqs, BSSortFlags;
Prototype const char    *BSortTempDir;
Prototype FILE        *(*BSortGetTempFile)(void *, unsigned int nr);
Prototype void          *BSortGetTempFilePtr;

/*********     GLOBAL VARIABLES                                 *********/
jsuint64_t BSortMaxMem = BIGSORT_MAXMEM;
unsigned int BSortNSeqs = 32; /*  at least 3  */
/*  bit 0: if set, truncate files to save  disk space (temporarily)
 */
unsigned int BSortFlags;
const char *BSortTempDir;
FILE *(*BSortGetTempFile)(void *, unsigned int nr);
void *BSortGetTempFilePtr;

#define getTempFile(nr)                          \
    (BSortGetTempFile == NULL ? getTempFile0(nr) : \
     BSortGetTempFile(BSortGetTempFilePtr, nr))

/* ERROR-DEFINITIONS from bigsortopen label _ERR_BIGSORTOPEN
   Too little memory for sorting
*/

/*JS*********************************************************************
*   Reads a stream until EOF is signaled and sorts the units of size len
*    using the given comparison function.
*    The routine 'doget' must return the length of the next unit, -1 on
*    error and -2 on EOF.
*    A handle is returned which allows direct or sequential access to the
*    sorted units. If no comparison routine is given (compare == NULL),
*    the stream is simply transferred to a direct-access one.
*    If nExpect >= 0, it states the number of entries expected.
*************************************************************************/

bigsorthandle_t *bigsortopen(bigsorthandle_t *bsh, unsigned int len,
                             bigsortcount_t nExpect,
                             int (*doget)(void *uptr, void *buff, unsigned int lbuff,
                                          bigsortcount_t nr),
                             int (*compare)(void *uptr,
                                            const void *a, unsigned int la,
                                            const void *b, unsigned int lb),
                             void *uPtr, int mode)

/************************************************************************/
{
    bigsortcount_t count;
    bigsortdata_t *buff;

    if((bsh = bigsortopen0(bsh, len, nExpect, compare, uPtr, mode)) == NULL)
        EGOTO(error0);

    if(doget == NULL)
        return bsh;

    if((buff =
#ifdef CONFIG_USE_ALLOCA
            alloca(len)
#else
            malloc(len)
#endif
        ) == NULL)
        EGOTO(error);

    for(count = 0 ; ; ++count) {
        int rc;

        if((rc = doget(uPtr, buff, len, count)) == -1)
            EGOTO(error);

        if(rc == -2)
            break;

        if(bigsortput(bsh, buff, rc) == (bigsortcount_t)-1)
            EGOTO(error);
    }

    /*  Finalize stream and do the actual sorting  */
    if(bigsortput(bsh, NULL, -1) == (bigsortcount_t)-1)
        EGOTO(error);

#ifndef CONFIG_USE_ALLOCA
    free(buff);
#endif

    return bsh;
error:
    (void)bigsortclose(bsh);

#ifndef CONFIG_USE_ALLOCA
    if(buff) free(buff);
#endif
error0:
    return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

static bigsorthandle_t *bigsortopen0(bigsorthandle_t *bsh, unsigned int len,
                                     bigsortcount_t nExpect,
                                     int (*compare)(void *uptr,
                                                    const void *a, unsigned int la,
                                                    const void *b, unsigned int lb),
                                     void *uPtr, int mode)

/************************************************************************/
{
    bigsort_t *pbs = NULL;
    size_t size0, size;
    bigsortcount_t maxEntries;
    unsigned int nSeqs, btype;

    CASSERT(sizeof(bigsortdata_t) == sizeof(char));

    /*  How many sequences may we use (must be at least 3)?  */
    nSeqs = BSortNSeqs;
    size0 = nSeqs * (sizeof(pbs->BSI_Array[0]) + len);
    if(nSeqs < 3 || size0 > BSortMaxMem) {
        nSeqs = 3;
        size0 = nSeqs * (sizeof(pbs->BSI_Array[0]) + len);
    }

    /*  Get maximum number of entries to be held in memory  */
    if(size0 >= BSortMaxMem)
        EGOTO(errmem);

    maxEntries = (BSortMaxMem - size0) / len;
    if(maxEntries < 2)
        EGOTO(errmem);

    /*  Use given number to restrict memory usage. This also works for
     *   nExpect == 0 or nExpect == -1 and does not produce invalid
     *   numbers!
     */
    if(nExpect > 0 && nExpect < maxEntries)
        maxEntries = nExpect;

#ifdef TRACETYPE_BIGSORT
    jstrace(TRACETYPE_BIGSORT, 1,
            "bigsortopen: using %d temporary files and %ld memory entries.\n",
            nSeqs, maxEntries);
#endif

    if((pbs = (bigsort_t *)malloc(sizeof(*pbs))) == NULL)
        EGOTO(error);

    /*  Set up internal structure  */
    pbs->BSI_Status  = BIGSORT_STATUS_DISTRUNSMEM;
    pbs->BSI_Mode    = mode;
    pbs->BSI_Len     = len;
    pbs->BSI_UPtr    = uPtr;
    pbs->BSI_Compare = compare;
    pbs->BSI_Seqs    = NULL;
    pbs->BSI_NSeqs   = nSeqs;
    pbs->BSI_Level   = 0;
    pbs->BSI_Curr    = 0;
    pbs->BSI_Count   = 0;
    pbs->BSI_NArray  = maxEntries;

    /*  Get (temporary) array  */
    assert(maxEntries > 0);
    size0 = maxEntries * sizeof(pbs->BSI_Array[0]);
    size0 = JS_ALIGN_OFFSET(size0, JSMAXALIGN);
    size = size0 + maxEntries * len;

    if((pbs->BSI_Array = (bigsortentry_t *)malloc(size)) == NULL)
        EGOTO(error);

    pbs->BSI_Array[0].BSIE_Data = (bigsortdata_t *)pbs->BSI_Array + size0;

    btype = BIGSORT_TYPE_TCOLLECTING;
    if(bsh != NULL) {
        btype |= BIGSORT_TYPE_FKEEPSTRUCT;

    } else if((bsh = (bigsorthandle_t *)malloc(sizeof(*bsh))) == NULL)
        EGOTO(error);

    pbs->BSI_Handle = bsh;

    bsh->BSH_Type  = btype;
    bsh->BSH_Len   = len;
    bsh->BSH_Curr  = 0;
    bsh->BSH_Count = 0;
    bsh->BSH_Buff  = NULL;
    bsh->BSH_Work  = pbs;
    bsh->BSH_Group = NULL;

    bsh->BSH_NTempFiles = bsh->BSH_MaxLevel = 0;
    bsh->BSH_TotalWritten = bsh->BSH_TotalRead = bsh->BSH_MaxTempSize = 0;
    bsh->BSH_TotalRuns = 0;

#ifdef CONFIG_BIGSORT_USEVARINT
    if(mode & BIGSORT_FVARLEN2)
        bsh->BSH_Type |= BIGSORT_TYPE_FVARLENGTH2;
    else
#endif
    if(mode & (BIGSORT_FVARLEN |BIGSORT_FVARLEN2))
        bsh->BSH_Type |= BIGSORT_TYPE_FVARLENGTH;

    if(mode & BIGSORT_FINVERTED)
        bsh->BSH_Type |= BIGSORT_TYPE_FINVERTED;

    if((mode & (BIGSORT_FGROUP | BIGSORT_FDISTINCT)) &&
       compare != NULL) {
        /*  Grouping or distinct requested  */
        if((bsh->BSH_Group = (bigsortgroup_t *)
            malloc(sizeof(*(bsh->BSH_Group)))) == NULL)
            EGOTO(error);

        bsh->BSH_Group->BSG_Previous = NULL;
        bsh->BSH_Group->BSG_PrevLen  = 0;

        bsh->BSH_Group->BSG_Compare = compare;
        bsh->BSH_Group->BSG_User = uPtr;
        if(mode & BIGSORT_FGROUP)
            bsh->BSH_Type |= BIGSORT_TYPE_FGROUPING;

    } else
        bsh->BSH_Group = NULL;

    return bsh;
errmem:
    JSSETERROR(_ERR_BIGSORTOPEN, 0);
error:
    if(pbs) {
        free(pbs->BSI_Array);
        free(pbs);
    }
    return NULL;
}

/*JS*********************************************************************
*   Closes the handle obtained by bigsortopen.
*************************************************************************/

int bigsortclose(bigsorthandle_t *bsh)

/************************************************************************/
{
    int ret;

    ret = 0;
    switch(bsh->BSH_Type & BIGSORT_TYPE_TMASK) {
    case BIGSORT_TYPE_TCOLLECTING: /*  still building up  */
	if(bsh->BSH_Work != NULL) {
	    bigsort_t *pbs;
	    bigsortseq_t *seqs;

	    pbs = (bigsort_t *)bsh->BSH_Work;

	    if((seqs = pbs->BSI_Seqs) != NULL) {
		unsigned int iseq;

		for(iseq = 0 ; iseq < pbs->BSI_NSeqs ; ++iseq)
		    if(seqs[iseq].SEQ_File != NULL &&
		       fclose(seqs[iseq].SEQ_File) == -1)
			ret = -1;

		free(pbs->BSI_Seqs);
	    }

	    if(pbs->BSI_Array) free(pbs->BSI_Array);

	    free(pbs);

	    bsh->BSH_Work = NULL;
	}

        break;

    case BIGSORT_TYPE_TMEMLIST: /*  memory list  */
        free(bsh->BSH_Buff);   bsh->BSH_Buff = NULL;
        break;

    case BIGSORT_TYPE_TFILELIST: /*  file access  */
        if(fclose(bsh->BSH_File) == EOF)
            ret = -1;

        bsh->BSH_File = NULL;
        break;

    default:
	assert(FALSE);
	break;
    }

    if(bsh->BSH_Group != NULL) {
        if(bsh->BSH_Group->BSG_Previous)
            free(bsh->BSH_Group->BSG_Previous);

        free(bsh->BSH_Group);
        bsh->BSH_Group = NULL;
    }

    if(!(bsh->BSH_Type & BIGSORT_TYPE_FKEEPSTRUCT))
        free(bsh);

    return ret;
}

/*JS*********************************************************************
*
*************************************************************************/

static int optimizeMemList(bigsorthandle_t *bsh, bigsortentry_t *array,
                           bigsortcount_t countold, bigsortcount_t countnew)

/************************************************************************/
{
    bigsortdata_t *from, *to;
    size_t sizeold, sizenew;
    ptrdiff_t diff;
    unsigned int len;

    if(countnew == 0) {
        free(array);
        bsh->BSH_Buff = NULL;
        return 0;
    }

    len = bsh->BSH_Len;

    sizeold = (size_t)countold * sizeof(array[0]);
    sizeold = JS_ALIGN_OFFSET(sizeold, JSMAXALIGN);

    sizenew = (size_t)countnew * sizeof(array[0]);
    sizenew = JS_ALIGN_OFFSET(sizenew, JSMAXALIGN);

    from = (bigsortdata_t *)array + sizeold;
    to   = (bigsortdata_t *)array + sizenew;
    if(to != from)
        memmove(to, from, countnew * len);

    /*  Minimize memory requirement  */
    if((bsh->BSH_Buff = (bigsortentry_t *)
        realloc(array, sizenew + (size_t)countnew * len)) == NULL)
        EGOTO(error);

    /*  We do not know if realloc to a smaller size may change the pointer  */
    if((diff = ((bigsortdata_t *)bsh->BSH_Buff - (bigsortdata_t *)array) + (to - from)) != 0) {
	array = bsh->BSH_Buff;

	while(--countnew != (bigsortcount_t)-1)
	    array[countnew].BSIE_Data = (bigsortdata_t *)array[countnew].BSIE_Data + diff;
    }

    return 0;
error:
    return -1;
}

/* ERROR-DEFINITIONS from bigsortput label _ERR_BIGSORTPUT
   Sorting stream already closed
*/

/*JS*********************************************************************
*
*************************************************************************/

bigsortcount_t bigsortput(bigsorthandle_t *bsh, const void *buff,
                          unsigned int lbuff)

/************************************************************************/
{
    bigsort_t *pbs;

    if((bsh->BSH_Type & BIGSORT_TYPE_TMASK) != BIGSORT_TYPE_TCOLLECTING) {
        JSSETERROR(_ERR_BIGSORTPUT, 0);
        EGOTO(error);
    }

    pbs = (bigsort_t *)bsh->BSH_Work;

    switch(pbs->BSI_Status) {
        /* ************************************************************* */
    case BIGSORT_STATUS_DISTRUNSMEM: /* ***  distribute runs in memory  *** */
        if(buff != NULL) {
            if(pbs->BSI_Count < pbs->BSI_NArray)
            {
                /*  Memory is still sufficient  */
                distElementMem(pbs, buff, lbuff);
            }
            else
            {
                /*  Internal array full, switch to external files  */
                if(setupExt(pbs) == -1)
                    EGOTO(error);

                if(distElementExt(pbs, buff, lbuff) == -1)
                    EGOTO(error);

                pbs->BSI_Status = BIGSORT_STATUS_DISTRUNSEXT;
            }

        } else { /*  user specified EOF, finalize stream  */
	    bigsortentry_t *array;
            bigsortcount_t nold, nnew;

            /* **  Simple case: All elements fit in memory, so sort and set up handle  ** */
	    array = pbs->BSI_Array;
	    nold  = pbs->BSI_NArray;
	    nnew = doQuickSort(pbs, pbs->BSI_Mode);

            bsh->BSH_Type  = (bsh->BSH_Type & ~BIGSORT_TYPE_TMASK) | BIGSORT_TYPE_TMEMLIST;
            bsh->BSH_Count = pbs->BSI_Count;

            free(pbs);

            if(optimizeMemList(bsh, array, nold, nnew) == -1)
                EGOTO(error);
        }

        break;

        /* ************************************************************* */
    case BIGSORT_STATUS_DISTRUNSEXT: /* ***  distribute runs in external storage (with comparison)  *** */
        if(buff != NULL) {
	    if(distElementExt(pbs, buff, lbuff) == -1)
		EGOTO(error);

        } else { /*  user specified EOF, finalize stream  */
            if(finalizeExt(pbs) == -1)
                EGOTO(error);

            /* **  Set up handle  ** */
            bsh->BSH_Type  = (pbs->BSI_Handle->BSH_Type & ~BIGSORT_TYPE_TMASK) | BIGSORT_TYPE_TFILELIST;
            bsh->BSH_Count = pbs->BSI_Count;
            bsh->BSH_File  = pbs->BSI_Seqs[0].SEQ_File; /*  this overwrites 'bsh->BSH_Work'  */

            /*  Now its time to free temporary work space from distribution  */
            free(pbs->BSI_Seqs);
            free(pbs);

            /*  We have to rewind the stream to make sure the file position
             *   corresponds with BSH_Curr. This is used implicitly in
             *   bigsortget.
             */
            if(fseek(bsh->BSH_File, 0, SEEK_SET) == -1)
                EGOTO(error);

            bsh->BSH_Type |= BIGSORT_TYPE_FFILEPOSOK;
        }

        break;

    default:
        assert(FALSE);
    }

    return 0;
error:
    return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static void distElementMem(bigsort_t *pbs, const void *buff,
                           unsigned int lbuff)

/************************************************************************/
{
    bigsortentry_t *array;
    bigsortcount_t count;
    unsigned int len;

    array = pbs->BSI_Array;
    len = pbs->BSI_Len;

    assert(lbuff <= len);

    count = pbs->BSI_Count++;

    assert(count < pbs->BSI_NArray);

    if(count > 0)
        array[count].BSIE_Data = (bigsortdata_t *)(array[count - 1].BSIE_Data) + len;

    /*  Store element  */
    memcpy(array[count].BSIE_Data, buff, lbuff);
    array[count].BSIE_Length = lbuff;
}

/*JS*********************************************************************
*
*************************************************************************/

static int setupExt(bigsort_t * pbs)

/************************************************************************/
{
    bigsortentry_t *array;
    bigsortcount_t nArray, ic;
    bigsortseq_t  *seqs;
    size_t size0, size;
    unsigned int iseq, len, nSeqs;

    array = pbs->BSI_Array;
    nArray = pbs->BSI_NArray;
    len = pbs->BSI_Len;

    assert(pbs->BSI_Count == nArray);

    nSeqs = pbs->BSI_Compare != NULL ? pbs->BSI_NSeqs : 1;

    size0 = nSeqs * sizeof(pbs->BSI_Seqs[0]);
    size0 = JS_ALIGN_OFFSET(size0, JSMAXALIGN);
    size  = size0 + nSeqs * len;

    if((seqs = (bigsortseq_t *)malloc(size)) == NULL)
        EGOTO(error);

    for(iseq = 0 ; iseq < nSeqs ; iseq++)
    {
        seqs[iseq].SEQ_File = NULL;
        seqs[iseq].SEQ_FileSize = 0;
        seqs[iseq].SEQ_Buffer = iseq == 0 ?
            (bigsortdata_t *)seqs + size0 : seqs[iseq - 1].SEQ_Buffer + len;

        seqs[iseq].SEQ_NRuns = seqs[iseq].SEQ_DiffRuns = iseq < (nSeqs - 1) ? 1 : 0;
        seqs[iseq].SEQ_Count = 0;
    }
    pbs->BSI_Seqs = seqs;
    pbs->BSI_NSeqs = nSeqs;

    if(pbs->BSI_Compare != NULL) {
        /* ***  Now initialize all streams  *** */
        if(selectRun(pbs) == -1)
            EGOTO(error);

        /* ***  Transfer array into a valid heap  *** */
        for(ic = nArray / 2 ; ic > 0 ; )
            bsSift(pbs, --ic, nArray);

        pbs->BSI_S = pbs->BSI_R = nArray;

#ifdef TRACETYPE_BIGSORT
        jstrace(TRACETYPE_BIGSORT, 2,
                "DIST %ld %ld\n", (long)pbs->BSI_S, (long)pbs->BSI_R);

        if(jstracelevel(TRACETYPE_BIGSORT) >= 8) {
            for(ic = 0 ; ic < pbs->BSI_R ; ic++)
                jstrace(TRACETYPE_BIGSORT, 8,
                        "  a[%ld]='%s'\n", (long)ic, array[ic].BSIE_Data);
        }
#endif
    }
    else
    {
        FILE *fp;

        /*  If no comparison routine given, simply copy everything to
         *   one stream and return
         */

        if((fp = getTempFile(0)) == NULL)
            EGOTO(error);

	pbs->BSI_Handle->BSH_NTempFiles++;

        pbs->BSI_Seqs[0].SEQ_File = fp;
        pbs->BSI_Seqs[0].SEQ_FileSize = 0;

        for(ic = 0; ic < nArray; ic++)
            if(storeContent(pbs->BSI_Handle, fp, &(pbs->BSI_Seqs[0].SEQ_FileSize),
                            array[ic].BSIE_Data, array[ic].BSIE_Length) == -1)
                EGOTO(error);
    }

    return 1;
error:
    return -1;
}

/*JS*********************************************************************
*  We maintain a lower and an upper heap in 'BSI_Array' (for the
*   explanation of "heap", see a documentation on the "heapsort"
*   algorithm).
*
*   The lower heap is in interval [0, BSI_S[ and the upper
*   in [BSI_S, BSI_R[.
*
*   The lowest element is in position "0" which is first
*   written to the temporary file.
*   If the next element is larger or equal than the lowest
*   element just written, is is still part of the current
*   "run" of elements and it is put in the lower heap:
*   copy it to position "0" and let it sink via "bsSift()"
*   to its proper position.
*   If the next element is smaller, it belongs to the
*   next run and it is added to the upper heap:
*   decrease lower heap by one (i. e. decrement BSI_S)
*   and move last element to first pos and maintain
*   heap property by "bsSift()". Put the new element
*   at start of upper heap (i. e. "BSI_S") and maintain
*   upper heap by "bsSift()" again.
*
*   Once the lower heap becomes empty, the upper heap
*   becomes the new lower heap (i. e. set BSI_S to BSI_R)
*
*   If no new elements are available (==EOF), we write
*   first the lower, then the upper heap in the right
*   sequence to the temporary file.
*************************************************************************/

static int distElementExt(bigsort_t *pbs, const void *buff,
                          unsigned int lbuff)

/************************************************************************/
{
    bigsortentry_t *array;
    bigsortseq_t *pseq;
    bigsortdata_t *data;
    unsigned int dlen;
    int (*compare)(void *uptr, const void *a, unsigned int la,
                   const void *b, unsigned int lb);
    void *uPtr;
    FILE *fp;

    array = pbs->BSI_Array;

    compare = pbs->BSI_Compare;
    uPtr = pbs->BSI_UPtr;

    pseq = &(pbs->BSI_Seqs[pbs->BSI_Curr]);
    fp = pseq->SEQ_File;

    if(compare == NULL) {
        assert(pbs->BSI_Curr == 0);
        assert(buff != NULL);

        /*  Write this element to stream  */
        data = (bigsortdata_t *)buff;
        dlen = lbuff;

    } else {
        /*  Write smallest element to stream  */
        data = array[0].BSIE_Data;
        dlen = array[0].BSIE_Length;
    }

    /* ********************************************************* */
    if(storeContent(pbs->BSI_Handle, fp, &(pseq->SEQ_FileSize), data, dlen) == -1)
        EGOTO(error);

    pseq->SEQ_Count++;

    if(compare == NULL) {
        pbs->BSI_Count++;
        return 0;
    }

    /* ********************************************************* */
    /*  "data" still contains the last written field  */
    if(buff != NULL) {
        int cmp;

        assert(pbs->BSI_R == pbs->BSI_NArray);

        assert(lbuff <= pbs->BSI_Len);

        pbs->BSI_Count++;

        cmp = compare(uPtr, data, dlen, buff, lbuff);
        if(pbs->BSI_Mode & BIGSORT_FINVERTED)
            cmp = -cmp;

        /*  add element *after* compare  */
        memcpy(data, buff, lbuff);
        dlen = lbuff;

        if(cmp <= 0) { /*  Element belongs to same run, put in lower heap  */
            array[0].BSIE_Length = dlen;

            bsSift(pbs, 0, pbs->BSI_S);
            return 0;
        }

    } else {
        /*  no more data to swallow, so shift upper heap left  */
        pbs->BSI_R--;

        /*   "pbs->BSI_S - 1 < pbs->BSI_R"  */
        if(pbs->BSI_S <= pbs->BSI_R) {
            data = array[pbs->BSI_R].BSIE_Data;
            dlen = array[pbs->BSI_R].BSIE_Length;

            array[pbs->BSI_R].BSIE_Data   = array[0].BSIE_Data;
            array[pbs->BSI_R].BSIE_Length = array[0].BSIE_Length;
        }
    }

    /* *****  Put element in upper heap  ***** */
    /*  (the new element is in "data" + "dlen")  */

    /*  we move element "S" -> "0" and the new element to "S"  */

    assert(pbs->BSI_S > 0);
    pbs->BSI_S--;
    if(pbs->BSI_S > 0) {
        array[0] = array[pbs->BSI_S];
        bsSift(pbs, 0, pbs->BSI_S);
    }

    /*  put the new element at position "S" (start of upper heap) and maintain upper heap  */
    array[pbs->BSI_S].BSIE_Data   = data;
    array[pbs->BSI_S].BSIE_Length = dlen;

    if(pbs->BSI_S < pbs->BSI_R && pbs->BSI_S < pbs->BSI_NArray / 2)
        bsSift(pbs, pbs->BSI_S, pbs->BSI_R);

    /*  This run finished  */
    if(pbs->BSI_S == 0) {
        /*  EOF reached?  */
        if(pbs->BSI_R == 0)
            return 1;

        /*  Start next run  */
        if(selectRun(pbs) == -1)
            EGOTO(error);

        pbs->BSI_S = pbs->BSI_R;
    }

    return 0;
error:
    return-1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int
finalizeExt(bigsort_t * pbs)

/************************************************************************/
{
    unsigned int iseq;
    int rc;

    if(pbs->BSI_Compare != NULL) {
        do {
            if((rc = distElementExt(pbs, NULL, -1)) == -1)
                EGOTO(error);

        } while(rc == 0);

        free(pbs->BSI_Array);   pbs->BSI_Array = NULL;

#ifdef TRACETYPE_BIGSORT
        if(jstracelevel(TRACETYPE_BIGSORT) >= 3) {
            for(iseq = 0 ; iseq < pbs->BSI_NSeqs ; ++iseq)
                jstrace(TRACETYPE_BIGSORT, 3,
                        "SEQ %d COUNT %ld NRUNS %ld DIFFRUNS %ld\n", iseq,
                        (long)pbs->BSI_Seqs[iseq].SEQ_Count,
                        (long)pbs->BSI_Seqs[iseq].SEQ_NRuns,
                        (long)pbs->BSI_Seqs[iseq].SEQ_DiffRuns);
        }
#endif

        if(mergeRuns(pbs) == -1)
            EGOTO(error);
    }
    else
    {
        free(pbs->BSI_Array);   pbs->BSI_Array = NULL;
    }

    rc = 0;
    /*  Close everything and clean up  */
    for(iseq = 1 ; iseq < pbs->BSI_NSeqs ; iseq++) {
        if(pbs->BSI_Seqs[iseq].SEQ_File == NULL)
	    continue;

	if(fclose(pbs->BSI_Seqs[iseq].SEQ_File) == -1)
            rc = -1;

	pbs->BSI_Seqs[iseq].SEQ_File = NULL;
    }

    if(rc == -1)
        EGOTO(error);

    return 0;
error:
    return -1;
}

/*JS*********************************************************************
*   Special case: Everything fits into memory, so we can do a quicksort
*    on a[0 .. n-1], if required.
*************************************************************************/

static bigsortcount_t doQuickSort(bigsort_t *pbs, int mode)

/************************************************************************/
{
    bigsortentry_t *array;
    bigsortcount_t r, s, n;
    int (*compare)(void *uptr, const void *a, unsigned int la,
                   const void *b, unsigned int lb);
    void *uPtr;

    array = pbs->BSI_Array;
    n     = pbs->BSI_Count;

    compare = pbs->BSI_Compare;
    uPtr = pbs->BSI_UPtr;

    if(compare != NULL) {
#ifdef CONFIG_BIGSORT_USEGENSORT
        if(mode & BIGSORT_FINVERTED)
            sortSpecialVarInv(array, n, compare, uPtr);
        else
            sortSpecialVar(array,    n, compare, uPtr);
#else
        CompareUPtr = uPtr;
        CompareFct  = compare;
        CompareLen  = pbs->BSI_Len;

        qsort(array, (size_t)n, sizeof(bigsortentry_t),
              mode & BIGSORT_FINVERTED ? sortCompareInv : sortCompare);
#endif

        if((mode & BIGSORT_FDISTINCT2) && n > 1) {
            /*  Remove duplicate entries  */
            for(s = 0, r = 1 ; r < n ; r++) {
                const bigsortdata_t *p1, *p2;
                unsigned int l1, l2;

                p1 = array[s].BSIE_Data;
                l1 = array[s].BSIE_Length;

                p2 = array[r].BSIE_Data;
                l2 = array[r].BSIE_Length;

                if(compare(uPtr, p1, l1, p2, l2) == 0)
                    continue;

                s++;
                if(s < r)
                    array[s] = array[r];
            }

	    ++s;

	    /*  set up new count after removing duplicates  */
            pbs->BSI_Count = s;

            /*  In case we did not align, we cannot reduce memory by setting
             *   n = s, since the sorting routine shuffled everything around.
             */
            if(mode & (BIGSORT_FVARLEN | BIGSORT_FVARLEN2))
                n = s;
        }
    }

    return n;
}

/*JS*********************************************************************
*   Sift routine, see a documentation for heapsort for an explanation.
*************************************************************************/

static void bsSift(bigsort_t *pbs, bigsortcount_t l, bigsortcount_t r)

/************************************************************************/
{
    bigsortentry_t *array;
    bigsortdata_t *ptemp;
    unsigned int ltemp;
    bigsortcount_t j;
    int (*compare)(void *uptr, const void *a, unsigned int la,
                   const void *b, unsigned int lb);
    void *uPtr;

    array = pbs->BSI_Array;

    compare = pbs->BSI_Compare;
    uPtr = pbs->BSI_UPtr;

    /*   Get first element to sink in heap   */
    ptemp = array[l].BSIE_Data;
    ltemp = array[l].BSIE_Length;

    while((j = (l << 1) + 1) < r) {
        const bigsortdata_t *p1;
        unsigned int l1;
        int cmp;

        p1 = array[j].BSIE_Data;
        l1 = array[j].BSIE_Length;

        if(j + 1 < r) {
            const bigsortdata_t *p2;
            unsigned int l2;

            p2 = array[j + 1].BSIE_Data;
            l2 = array[j + 1].BSIE_Length;

            cmp = compare(uPtr, p1, l1, p2, l2);
            if(pbs->BSI_Mode & BIGSORT_FINVERTED)
                cmp = -cmp;

            if(cmp > 0) {
                j++;
                p1 = p2;
                l1 = l2;
            }
        }

        cmp = compare(uPtr, ptemp, ltemp, p1, l1);
        if(pbs->BSI_Mode & BIGSORT_FINVERTED)
            cmp = -cmp;

        if(cmp <= 0)
            break;

        /*   Copy lower element to top   */
        array[l] = array[j];

        l = j;
    }

    /*   Now the first element has reached its place   */
    array[l].BSIE_Data   = ptemp;
    array[l].BSIE_Length = ltemp;
}

/*JS*********************************************************************
*   Sift routine, see a documentation for heapsort for an explanation.
*************************************************************************/

static void bsSiftInd(bigsort_t *pbs, bigsortcount_t l, bigsortcount_t r,
                      unsigned int *tr)

/************************************************************************/
{
    bigsortseq_t  *seqs;
    bigsortdata_t *ptemp;
    unsigned int ltemp, itemp;
    bigsortcount_t j;
    int (*compare)(void *uptr, const void *a, unsigned int la,
                   const void *b, unsigned int lb);
    void *uPtr;

    seqs  = pbs->BSI_Seqs;

    compare = pbs->BSI_Compare;
    uPtr = pbs->BSI_UPtr;

    /*   Get first element to sink in heap   */
    itemp = tr[l];
    ptemp = seqs[itemp].SEQ_Buffer;
    ltemp = seqs[itemp].SEQ_BuffLen;

    while((j = (l << 1) + 1) < r) {
        const bigsortdata_t *p1;
        unsigned int l1;
        int cmp;

        p1 = seqs[tr[j]].SEQ_Buffer;
        l1 = seqs[tr[j]].SEQ_BuffLen;

        if(j + 1 < r) {
            const bigsortdata_t *p2;
            unsigned int l2;

            p2 = seqs[tr[j + 1]].SEQ_Buffer;
            l2 = seqs[tr[j + 1]].SEQ_BuffLen;

            cmp = compare(uPtr, p1, l1, p2, l2);
            if(pbs->BSI_Mode & BIGSORT_FINVERTED)
                cmp = -cmp;

            if(cmp > 0) {
                j++;
                p1 = p2;
                l1 = l2;
            }
        }

        cmp = compare(uPtr, ptemp, ltemp, p1, l1);
        if(pbs->BSI_Mode & BIGSORT_FINVERTED)
            cmp = -cmp;

        if(cmp <= 0)
            break;

        /*   Copy lower element to top   */
        tr[l] = tr[j];

        l = j;
    }

    /*   Now the first element has reached its place   */
    tr[l] = itemp;
}

#ifdef CONFIG_BIGSORT_USEGENSORT
/*JS*********************************************************************
*   If the stream fits completely into memory, use a quicksort to sort it.
*************************************************************************/

/*  Generated automatically with "gensort.h"  */
#define GENSORT_NAME                 sortSpecialVar
#define GENSORT_INTTYPE              bigsortcount_t
#define GENSORT_ARGS                 , compare, uPtr
#define GENSORT_ARGSPROTO            , int (*compare)(void *uptr,       \
                                                      const void *a, unsigned int la, \
                                                      const void *b, unsigned int lb), \
                                       void *uPtr
#define GENSORT_TYPE                 bigsortentry_t
#define GENSORT_KEYTYPE              bigsortentry_t
#define GENSORT_GETKEY(a)            a
#define GENSORT_COMPAREKEYS(k1, k2)  compare(uPtr, k1.BSIE_Data, k1.BSIE_Length, \
                                             k2.BSIE_Data, k2.BSIE_Length) < 0
#define GENSORT_STATIC
#define GENSORT_NOCAST_GETKEY
#include <gensort.h>

/*  Generated automatically with "gensort.h"  */
#define GENSORT_NAME                 sortSpecialVarInv
#define GENSORT_INTTYPE              bigsortcount_t
#define GENSORT_ARGS                 , compare, uPtr
#define GENSORT_ARGSPROTO            , int (*compare)(void *uptr,       \
                                                      const void *a, unsigned int la, \
                                                      const void *b, unsigned int lb), \
                                       void *uPtr
#define GENSORT_TYPE                 bigsortentry_t
#define GENSORT_KEYTYPE              bigsortentry_t
#define GENSORT_GETKEY(a)            a
#define GENSORT_COMPAREKEYS(k1, k2)  compare(uPtr, k1.BSIE_Data, k1.BSIE_Length, \
                                             k2.BSIE_Data, k2.BSIE_Length) > 0
#define GENSORT_STATIC
#define GENSORT_NOCAST_GETKEY

#include <gensort.h>

#else

static void *CompareUPtr;
static int (*CompareFct)(void *uptr, const void *a, unsigned int la,
                         const void *b, unsigned int lb);
static unsigned int CompareLen;

/*JS*********************************************************************
*   If the stream fits completely into memory, use a quicksort to sort it.
*************************************************************************/

static int sortCompare(const void *a, const void *b)
{
    const bigsortentry_t *pa, *pb;

    pa = (const bigsortentry_t *)a;
    pb = (const bigsortentry_t *)b;

    return CompareFct(CompareUPtr, pa->BSIE_Data, pa->BSIE_Length, pb->BSIE_Data, pb->BSIE_Length);
}

static int sortCompareInv(const void *a, const void *b)
{
    const bigsortentry_t *pa, *pb;

    pa = (const bigsortentry_t *)a;
    pb = (const bigsortentry_t *)b;

    return -CompareFct(CompareUPtr, pa->BSIE_Data, pa->BSIE_Length, pb->BSIE_Data, pb->BSIE_Length);
}
#endif

/*JS*********************************************************************
*   Select the next sequence to copy the next run to.
*************************************************************************/

static int selectRun(bigsort_t *pbs)

/************************************************************************/
{
    unsigned int curr;

    curr = pbs->BSI_Curr;
    if(pbs->BSI_Seqs[curr].SEQ_DiffRuns <
       pbs->BSI_Seqs[curr + 1].SEQ_DiffRuns) {
        pbs->BSI_Curr = ++curr;

    } else if(pbs->BSI_Seqs[curr].SEQ_DiffRuns == 0) {
        bigsortcount_t nRuns0;
        unsigned int i;

        pbs->BSI_Level++;

        /*
         *  Pseudo runs are exhausted on all levels, so step up one level.
         *
         *  Here we use the recursion relation of the Fibonacci number of
         *   the order 'BSI_NSeqs' to calculate 'SEQ_NRuns_(n+1)' from
         *   'SEQ_NRuns_n'
         */
        nRuns0 = pbs->BSI_Seqs[0].SEQ_NRuns;
        for(i = 0 ; i < (pbs->BSI_NSeqs - 1) ; i++) {
            bigsortcount_t nruns_old;

            /*  all pseudo runs are exhausted  */
            assert(pbs->BSI_Seqs[i].SEQ_DiffRuns == 0);

            nruns_old = pbs->BSI_Seqs[i].SEQ_NRuns;

            /*  recursion relation  */
            pbs->BSI_Seqs[i].SEQ_NRuns = nRuns0 + pbs->BSI_Seqs[i + 1].SEQ_NRuns;

            /*  Calculate the pseudo runs contained  */
            pbs->BSI_Seqs[i].SEQ_DiffRuns = pbs->BSI_Seqs[i].SEQ_NRuns - nruns_old;
        }

        pbs->BSI_Curr = curr = 0;

    } else
        pbs->BSI_Curr = curr = 0;

    /*  substitute one pseudo run with a real one  */
    assert(pbs->BSI_Seqs[curr].SEQ_DiffRuns > 0);
    pbs->BSI_Seqs[curr].SEQ_DiffRuns--;

    pbs->BSI_Handle->BSH_TotalRuns++;

    if(pbs->BSI_Seqs[curr].SEQ_File == NULL) {
	if((pbs->BSI_Seqs[curr].SEQ_File = getTempFile(curr)) == NULL)
	    EGOTO(error);

	pbs->BSI_Seqs[curr].SEQ_FileSize = 0;
	pbs->BSI_Handle->BSH_NTempFiles++;
    }

    return 0;
error:
    return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int
storeContent(bigsorthandle_t *bsh, FILE * fp, bigsortcount_t *psize,
	     const bigsortdata_t *data, bigsortvarlen_t dlen)

/************************************************************************/
{
    unsigned int off;

#ifdef CONFIG_BIGSORT_USEVARINT
    if(bsh->BSH_Type & BIGSORT_TYPE_FVARLENGTH2)
    {
        bigsortdata_t buffer[JSVARINT_MAXLEN + 1], *pbuffer;

        pbuffer = buffer;
        JSVARINTSTORE32(pbuffer, dlen);

        off = pbuffer - buffer;

        if(fwrite(buffer, off, 1, fp) != 1)
            EGOTO(error);
    }
    else
#endif
    if(bsh->BSH_Type & BIGSORT_TYPE_FVARLENGTH)
    {
        if(fwrite(&dlen, sizeof(dlen), 1, fp) != 1)
            EGOTO(error);

        off = sizeof(dlen);
    }
    else
    {
        assert(dlen == bsh->BSH_Len);
        off = 0;
    }

    if(dlen > 0 && fwrite(data, dlen, 1, fp) != 1)
        EGOTO(error);

    off += dlen;

    bsh->BSH_TotalWritten += off;

    if(psize)
        *psize += off;

    return off;
error:
    return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int
readContent(bigsorthandle_t *bsh, FILE * fp, bigsortdata_t *data,
            bigsortvarlen_t *pdlen)

/************************************************************************/
{
    bigsortvarlen_t dlen;
    unsigned int off;

#ifdef CONFIG_BIGSORT_USEVARINT
    if(bsh->BSH_Type & BIGSORT_TYPE_FVARLENGTH2)
    {
        bigsortdata_t buffer[JSVARINT_MAXLEN + 1], *pbuffer;

        /*  a variable integer extends as long as the high bit is set  */
        for(off = 0; off < JSVARINT_MAXLEN;)
        {
            int c;

            if((c = getc(fp)) == EOF)
                EGOTO(error);

            buffer[off++] = (bigsortdata_t)c;

            if(!(c & 0x80))
                break;
        }

        if(off == JSVARINT_MAXLEN)
            EGOTO(error);

        pbuffer = buffer;
        dlen = JSVARINTREAD32(pbuffer);
        assert(pbuffer == (buffer + off));
    }
    else
#endif
    if(bsh->BSH_Type & BIGSORT_TYPE_FVARLENGTH)
    {
        if(fread(&dlen, sizeof(dlen), 1, fp) != 1)
            EGOTO(error);

        off = sizeof(dlen);

        assert(dlen <= bsh->BSH_Len);
    }
    else
    {
        dlen = bsh->BSH_Len;
        off = 0;
    }

    if(dlen > 0 && fread(data, dlen, 1, fp) != 1)
        EGOTO(error);

    off += dlen;

    *pdlen = dlen;

    bsh->BSH_TotalRead += off;

    return off;
error:
    return -1;
}

/*JS*********************************************************************
*   The I/O intensive sorting is done here: Merge runs on nSeqs-1 input
*    streams to one output stream until one stream is exhausted. Then
*    exchange the output stream with the empty input stream and start
*    over. The number of runs on the input stream (i.e. real ones and
*    pseudo runs) must fulfill the condition of the Fibonacci number of
*    a certain order.
*************************************************************************/

static int mergeRuns(bigsort_t *pbs)

/************************************************************************/
{
    unsigned int *tr, i, nSeqs;

    nSeqs = pbs->BSI_NSeqs;
    assert(nSeqs >= 3);

    if((tr = (unsigned int *)
#ifdef CONFIG_USE_ALLOCA
        alloca(nSeqs * sizeof(*tr))
#else
        malloc(nSeqs * sizeof(*tr))
#endif
           ) == NULL)
        EGOTO(error);

    /*  Initialize reading for input sequences  */
    for(i = 0 ; i < (nSeqs - 1) ; i++) {
        FILE *fp;

        if(pbs->BSI_Seqs[i].SEQ_Count == 0)
            continue;

        fp = pbs->BSI_Seqs[i].SEQ_File;

        if(fseek(fp, 0L, SEEK_SET) == -1)
            EGOTO(error);

        if(readContent(pbs->BSI_Handle, fp, pbs->BSI_Seqs[i].SEQ_Buffer,
                       &(pbs->BSI_Seqs[i].SEQ_BuffLen)) == -1)
            EGOTO(error);
    }

    /*  Create missing file for merging  */
    if((pbs->BSI_Seqs[nSeqs - 1].SEQ_File = getTempFile(nSeqs - 1)) == NULL)
        EGOTO(error);

    pbs->BSI_Seqs[nSeqs - 1].SEQ_FileSize = 0;
    pbs->BSI_Handle->BSH_NTempFiles++;

    /*  memorize max level  */
    pbs->BSI_Handle->BSH_MaxLevel = pbs->BSI_Level;

    do {
        FILE *fpdest;
        bigsortcount_t *psizedest, remRuns;
        unsigned int k;

        /* ***  Mix required number of runs for this level from sequences
         * ***   0 .. nSeqs-2 to output sequence nSeqs-1.
         */

        /*  Clear output sequence  */
        fpdest = pbs->BSI_Seqs[nSeqs - 1].SEQ_File;
	psizedest = &(pbs->BSI_Seqs[nSeqs - 1].SEQ_FileSize);

        if(fseek(fpdest, 0L, SEEK_SET) == -1)
            EGOTO(error);

#ifdef CONFIG_BIGSORT_USETRUNCATE
        /*  Truncate the output file to zero length  */
	if((BSortFlags & 1) && ftruncate(fileno(fpdest), 0) == -1)
            EGOTO(error);
#endif
	*psizedest = 0;

        pbs->BSI_Seqs[nSeqs - 1].SEQ_DiffRuns = 0;
        pbs->BSI_Seqs[nSeqs - 1].SEQ_Count    = 0;

#ifdef TRACETYPE_BIGSORT
        if(jstracelevel(TRACETYPE_BIGSORT) >= 2) {
            bigsortcount_t sum;

            jstrace(TRACETYPE_BIGSORT, 2, "Level %d\n", pbs->BSI_Level);
            sum = 0;
            for(i = 0 ; i < nSeqs ; i++) {
                if(pbs->BSI_Seqs[i].SEQ_Count)
                    jstrace(TRACETYPE_BIGSORT, 2,
                            "  Seq[%d]: nruns %ld diffruns %ld count %ld\n",
                            i,
                            (long)pbs->BSI_Seqs[i].SEQ_NRuns,
                            (long)pbs->BSI_Seqs[i].SEQ_DiffRuns,
                            (long)pbs->BSI_Seqs[i].SEQ_Count);

                sum += pbs->BSI_Seqs[i].SEQ_Count;
            }

            jstrace(TRACETYPE_BIGSORT, 2, "%ld on sequences.\n", (long)sum);
        }
#endif

        for(remRuns = pbs->BSI_Seqs[nSeqs - 2].SEQ_NRuns ;
            remRuns > 0 ;
            remRuns--) {
            /*  Count number of active sequences  */
            for(k = i = 0 ; i < (nSeqs - 1) ; i++) {
                assert(pbs->BSI_Seqs[i].SEQ_NRuns >= pbs->BSI_Seqs[i].SEQ_DiffRuns);

                /*  We might have SEQ_NRuns > SEQ_DiffRuns, but still SEQ_Count == 0 since
                 *   two sequences on this stream might have become one
                 */
                assert(pbs->BSI_Seqs[i].SEQ_NRuns > pbs->BSI_Seqs[i].SEQ_DiffRuns ||
                       pbs->BSI_Seqs[i].SEQ_Count == 0);

                if(pbs->BSI_Seqs[i].SEQ_DiffRuns > 0) {
                    /*  Pseudo run on sequence i  */
                    pbs->BSI_Seqs[i].SEQ_DiffRuns--;

                } else if(pbs->BSI_Seqs[i].SEQ_Count > 0) {
		    assert(pbs->BSI_Seqs[i].SEQ_NRuns > 0);
                    tr[k++] = i;
                }
            }

#ifdef TRACETYPE_BIGSORT
            if(k > 0)
                jstrace(TRACETYPE_BIGSORT, 2, "Have %d active sequences.\n", k);
#endif

            if(k == 0) {
                /*  All sequences have at least one pseudo run, so
                 *   record one pseudo run for destination.
                 */
                pbs->BSI_Seqs[nSeqs - 1].SEQ_DiffRuns++;
                continue;
            }

#if 0
            printf("NSEQ %d\n", nSeqs);
            for(i = 0 ; i < k ; ++i)
		printf("%d:%d %d\n", i, tr[i], pbs->BSI_Seqs[i].SEQ_Count);
#endif

            /*  Convert array tr into a heap to be able to locate the minimum element easily  */
            if(k > 2) {
                for(i = k / 2 - 1 ; i != (unsigned int)-1 ; --i)
                    bsSiftInd(pbs, i, k, tr);
            }

            /* **  Now copy one run from sequences t[0] .. t[k-1] to nSeqs-1 ** */
            do {
                bigsortdata_t *pmi;
                unsigned int lmi, tr_mi;

                /*  Find minimal element  */
                bsSiftInd(pbs, 0, k, tr);

                /*  Write minimal element from sequence tr[mi] to sequence nSeqs-1  */
                tr_mi = tr[0];
#if 0 /*  check minimality  */
		for(i = 1 ; i < k ; ++i)
		    assert(pbs->BSI_Compare(pbs->BSI_UPtr,
					    pbs->BSI_Seqs[tr_mi].SEQ_Buffer,
					    pbs->BSI_Seqs[tr_mi].SEQ_BuffLen,
					    pbs->BSI_Seqs[tr[i]].SEQ_Buffer,
					    pbs->BSI_Seqs[tr[i]].SEQ_BuffLen) <= 0);
#endif


                pmi = pbs->BSI_Seqs[tr_mi].SEQ_Buffer;
                lmi = pbs->BSI_Seqs[tr_mi].SEQ_BuffLen;
                pbs->BSI_Seqs[tr_mi].SEQ_Buffer  = pbs->BSI_Seqs[nSeqs - 1].SEQ_Buffer;
                pbs->BSI_Seqs[tr_mi].SEQ_BuffLen = pbs->BSI_Seqs[nSeqs - 1].SEQ_BuffLen;
                pbs->BSI_Seqs[nSeqs - 1].SEQ_Buffer  = pmi;
                pbs->BSI_Seqs[nSeqs - 1].SEQ_BuffLen = lmi;

                if(storeContent(pbs->BSI_Handle, fpdest, psizedest, pmi, lmi) == -1)
                    EGOTO(error);

#ifdef TRACETYPE_BIGSORT
                jstrace(TRACETYPE_BIGSORT, 8,
                        "%d %d %d: %ld %ld\n", k,
                        nSeqs - 1, tr_mi,
                        (long)pbs->BSI_Seqs[nSeqs - 1].SEQ_Count,
                        (long)pbs->BSI_Seqs[tr_mi].SEQ_Count);
#endif

                pbs->BSI_Seqs[nSeqs - 1].SEQ_Count++;

                assert(pbs->BSI_Seqs[tr_mi].SEQ_Count > 0);
                pbs->BSI_Seqs[tr_mi].SEQ_Count--;

                /*  Read next element from input sequence, check for EOF  */
                if(pbs->BSI_Seqs[tr_mi].SEQ_Count > 0) {
                    bigsortdata_t *ptmp;
                    bigsortvarlen_t ltmp;
                    FILE *fp;
                    int cmp;

                    fp = pbs->BSI_Seqs[tr_mi].SEQ_File;

                    ptmp = pbs->BSI_Seqs[tr_mi].SEQ_Buffer;

                    if(readContent(pbs->BSI_Handle, fp, ptmp, &ltmp) == -1)
                        EGOTO(error);

                    pbs->BSI_Seqs[tr_mi].SEQ_BuffLen = ltmp;

                    /*  Check for end of run  */
                    cmp = pbs->BSI_Compare(pbs->BSI_UPtr, pmi, lmi, ptmp, ltmp);
                    if(pbs->BSI_Mode & BIGSORT_FINVERTED)
                        cmp = -cmp;

                    if(cmp <= 0)
                        continue;
                }

#ifdef TRACETYPE_BIGSORT
                jstrace(TRACETYPE_BIGSORT, 4, "%s on sequence %d after %ld entries.\n",
                        pbs->BSI_Seqs[tr_mi].SEQ_Count == 0 ?
                        "End of file" : "End of run",
			tr_mi, pbs->BSI_Seqs[nSeqs - 1].SEQ_Count);
#endif

                /*  Remove this sequence from input sequences  */
                tr[0] = tr[--k];

            } while(k > 0);
        }

	/* ***  Determine maximum size of temporary files  *** */
	{
	    bigsortcount_t totalTempSize;

	    totalTempSize = 0;
	    for(i = 0 ; i < nSeqs ; ++i)
	        totalTempSize += pbs->BSI_Seqs[i].SEQ_FileSize;

	    if(totalTempSize > pbs->BSI_Handle->BSH_MaxTempSize)
	        pbs->BSI_Handle->BSH_MaxTempSize = totalTempSize;
	}

        /* ***  Rotate sequences  *** */
#ifdef TRACETYPE_BIGSORT
        jstrace(TRACETYPE_BIGSORT, 4, "Rotating sequences\n");
#endif
        {
            bigsortseq_t save;
            bigsortcount_t diff;

            save = pbs->BSI_Seqs[nSeqs - 1];
            diff = pbs->BSI_Seqs[nSeqs - 2].SEQ_NRuns;
            for(i = nSeqs - 1 ; i > 0 ; i--) {
                pbs->BSI_Seqs[i] = pbs->BSI_Seqs[i - 1];
                pbs->BSI_Seqs[i].SEQ_NRuns -= diff;
            }

            pbs->BSI_Seqs[0] = save;
            pbs->BSI_Seqs[0].SEQ_NRuns = diff;
        }

        assert(pbs->BSI_Seqs[nSeqs - 1].SEQ_NRuns == 0 &&
               pbs->BSI_Seqs[nSeqs - 1].SEQ_Count == 0);

        /* ***  Initialize read on former output sequence  *** */
        fpdest = pbs->BSI_Seqs[0].SEQ_File;
        if(fseek(fpdest, 0L, SEEK_SET) == -1)
            EGOTO(error);

        if(readContent(pbs->BSI_Handle, fpdest, pbs->BSI_Seqs[0].SEQ_Buffer,
                       &(pbs->BSI_Seqs[0].SEQ_BuffLen)) == -1)
            EGOTO(error);

        /*  We have finally decreased the level by one  */
        pbs->BSI_Level--;

    } while(pbs->BSI_Level != (unsigned int)-1);

#ifndef NDEBUG
    /*  If everything worked, all items are now in sequence 0
     *   and the rest is empty.
     */
    assert(pbs->BSI_Seqs[0].SEQ_Count == pbs->BSI_Count);
    for(i = 1 ; i < nSeqs ; ++i) {
        assert(pbs->BSI_Seqs[i].SEQ_Count == 0);
        assert(pbs->BSI_Seqs[i].SEQ_NRuns == 0);
        assert(pbs->BSI_Seqs[i].SEQ_DiffRuns == 0);
    }
#endif

    /*  Remove duplicate entries on stream  */
    if((pbs->BSI_Mode & BIGSORT_FDISTINCT2) &&
       removeduplicates(pbs) == -1)
        EGOTO(error);

    /*  Rewind resulting sequence  */
    if(fseek(pbs->BSI_Seqs[0].SEQ_File, 0L, SEEK_SET) == -1)
        EGOTO(error);

#ifndef CONFIG_USE_ALLOCA
    free(tr);
#endif

    return 0;
error:
#ifndef CONFIG_USE_ALLOCA
    if(tr)
        free(tr);
#endif
    return -1;
}

/*JS*********************************************************************
*   Remove duplicate entries in the file.
*************************************************************************/

static int removeduplicates(bigsort_t *pbs)

/************************************************************************/
{
    FILE *fp;
    bigsortdata_t *sbuff, *rbuff;
    bigsortvarlen_t slen, slenf;
    off_t soff, roff;
    bigsortcount_t s, r, n;

    fp = pbs->BSI_Seqs[0].SEQ_File;
    sbuff = pbs->BSI_Seqs[0].SEQ_Buffer;
    rbuff = pbs->BSI_Seqs[1].SEQ_Buffer;
    n = pbs->BSI_Count;

    soff = 0;
    if(fseek(fp, soff, SEEK_SET) == -1)
        EGOTO(error);

    if((slenf = readContent(pbs->BSI_Handle, fp, sbuff, &slen)) == (unsigned int)-1)
        EGOTO(error);

    roff = slenf;
    for(s = 0, r = 1 ; r < n ; ++r) {
        bigsortvarlen_t rlen, rlenf;

        assert(ftell(fp) == roff);

        if((rlenf = readContent(pbs->BSI_Handle, fp, rbuff, &rlen)) == (unsigned int)-1)
            EGOTO(error);

        roff += rlenf;

        if(pbs->BSI_Compare(pbs->BSI_UPtr, sbuff, slen, rbuff, rlen) == 0)
            continue;

        /*  Go to next element in memory and in file  */
        ++s;
        soff += slenf;
        { bigsortdata_t *tmp = sbuff; sbuff = rbuff; sbuff = tmp; }
        slen = rlen;
        slenf = rlenf;

        if(s < r) {
            if(fseek(fp, soff, SEEK_SET) == -1)
                EGOTO(error);

            if(storeContent(pbs->BSI_Handle, fp, NULL, sbuff, slen) == -1)
                EGOTO(error);

            if(fseek(fp, roff, SEEK_SET) == -1)
                EGOTO(error);
        }
    }

    pbs->BSI_Count = ++s;

#ifndef CONFIG_NO_POSIX
    /*  Truncate the output file to save disk space  */
    if((BSortFlags & 1) && ftruncate(fileno(fp), soff) == -1)
        EGOTO(error);
#endif

    return 0;
error:
    return -1;
}

/*JS*********************************************************************
*   Returns a temporary file. May be changed to use a different directory
*    for temporary files.
*************************************************************************/

static FILE *getTempFile0(unsigned int nr)

/************************************************************************/
{
    FILE *fp;

    if(BSortTempDir) {
        char name[JS_FILENAME_MAX + 1];

#ifndef CONFIG_NO_POSIX
        sprintf(name, "%s/bstp%d_%01d", BSortTempDir, (int)getpid(), nr);
#else
        sprintf(name, "%s/bstmp%02d", BSortTempDir, nr);
#endif

        if((fp = fopen(name, "wb+")) == NULL)
            EGOTO(error);

#ifdef TRACETYPE_BIGSORT
        /*  Maybe not delete them?  */
        if(jstracelevel(TRACETYPE_BIGSORT) >= 8)
            jstrace(TRACETYPE_BIGSORT, 8, "TEMPFILE '%s'\n", name);
        else
#endif
        if(remove(name) < 0) {
            fclose(fp);
            EGOTO(error);
        }
    } else
        fp = tmpfile();

    return fp;
error:
    return NULL;
}

/* ERROR-DEFINITIONS from bigsortget label _ERR_BIGSORTGET
   Sorting stream not yet closed
   No direct access for variable length stream
   Buffer too small
*/

/*JS*********************************************************************
*   Provides direct access to the sorted stream. Read set nr (0 .. n-1)
*    into buff and return length in buffer read.
*    On error -1 is returned and -2 on EOF.
*    If nr == -1, the next set in the sequence is returned (sequential
*    access). nr == -2 is similar to nr == -1, but the current position
*    is not moved.
*    If the handle does grouping, the routine returns (rlen << 1) | 1
*    if the expression changed and (rlen << 1) | 0 if not.
*************************************************************************/

int bigsortget(bigsorthandle_t *bsh, bigsortcount_t nr,
               void *buff, unsigned int lbuff)

/************************************************************************/
{
    bigsortcount_t curr;
    bigsortvarlen_t blen;
    unsigned int len;
    int rc;

    if((bsh->BSH_Type & BIGSORT_TYPE_TMASK) == BIGSORT_TYPE_TCOLLECTING &&
       /*  Do forgotten finalization  */
       bigsortput(bsh, NULL, -1) == (bigsortcount_t)-1)
        EGOTO(error);

    /*  No direct access when grouping or removing duplicates  */
    if(bsh->BSH_Group && nr != (bigsortcount_t)-1 && nr != (bigsortcount_t)-2)
        EGOTO(error);

    len = bsh->BSH_Len;

    /*  Current position  */
    curr = (nr != (bigsortcount_t)-1 && nr != (bigsortcount_t)-2) ? nr : bsh->BSH_Curr;

 restart:
    /*  Check for EOF  */
    if(curr >= bsh->BSH_Count)
        return -2; /*  EOF  */

    switch(bsh->BSH_Type & BIGSORT_TYPE_TMASK) {
    case BIGSORT_TYPE_TMEMLIST:
        {
            const bigsortdata_t *ptr;

            ptr = bsh->BSH_Buff[curr].BSIE_Data;
            blen = bsh->BSH_Buff[curr].BSIE_Length;

            if(blen > lbuff) {
                JSSETERROR(_ERR_BIGSORTGET, 2);
                EGOTO(error);
            }

            memcpy(buff, ptr, blen);

            rc = blen;
            if(nr != (bigsortcount_t)-2)
                bsh->BSH_Curr = curr + 1;

            if(bsh->BSH_Group != NULL) {
                rc <<= 1;

                if(curr == 0) {
                    rc |= 1;

                } else {
                    const bigsortdata_t *pprev;
                    unsigned int lprev;

                    pprev = bsh->BSH_Buff[curr - 1].BSIE_Data;
                    lprev = bsh->BSH_Buff[curr - 1].BSIE_Length;

                    if(bsh->BSH_Group->BSG_Compare(bsh->BSH_Group->BSG_User,
                                                   pprev, lprev, buff, blen) != 0) {
                        rc |= 1;

                    } else if(!(bsh->BSH_Type & BIGSORT_TYPE_FGROUPING)) {
                        ++curr;
                        goto restart;
                    }

                }
            }
        }
        break;

    case BIGSORT_TYPE_TFILELIST:
        if((curr != bsh->BSH_Curr ||
            !(bsh->BSH_Type & BIGSORT_TYPE_FFILEPOSOK))) {
            if(bsh->BSH_Type & (BIGSORT_TYPE_FVARLENGTH | BIGSORT_TYPE_FVARLENGTH2)) {
                JSSETERROR(_ERR_BIGSORTGET, 1);
                EGOTO(error);
            }

            if(fseek(bsh->BSH_File, (off_t)curr * len, SEEK_SET) == -1)
                EGOTO(error);
        }

        if(readContent(bsh, bsh->BSH_File, (bigsortdata_t *)buff, &blen) == -1)
            EGOTO(error);

        if(blen > lbuff) {
            JSSETERROR(_ERR_BIGSORTGET, 2);
            EGOTO(error);
        }

        rc = blen;
        if(nr != (bigsortcount_t)-2) {
            bsh->BSH_Curr = curr + 1;
            /*  BSH_Curr again matches filepos  */
            bsh->BSH_Type |= BIGSORT_TYPE_FFILEPOSOK;

        } else
            bsh->BSH_Type &= ~BIGSORT_TYPE_FFILEPOSOK;

        if(bsh->BSH_Group) {
            bigsortdata_t *pprev;
            unsigned int lprev;

            rc <<= 1;

            if((pprev = bsh->BSH_Group->BSG_Previous) == NULL) {
                if((pprev = (bigsortdata_t *)malloc(len)) == NULL)
                    EGOTO(error);

                bsh->BSH_Group->BSG_Previous = pprev;
                bsh->BSH_Group->BSG_PrevLen  = lprev = 0;
            }
            else
                lprev = bsh->BSH_Group->BSG_PrevLen;

            /*  Stream might have been rewound, so we check curr for start  */
            if(curr == 0) {
                rc |= 1;

            } else if(bsh->BSH_Group->BSG_Compare(bsh->BSH_Group->BSG_User,
                                                  pprev, lprev, buff, blen) != 0) {
                rc |= 1;

            } else if(!(bsh->BSH_Type & BIGSORT_TYPE_FGROUPING)) {
                ++curr;
                goto restart;
            }

            memcpy(pprev, buff, blen);
            bsh->BSH_Group->BSG_PrevLen = lprev;
        }

        break;

    default:
        assert(FALSE);
    }

    return rc;
error:
    return -1;
}

/*JS*********************************************************************
*   Provides direct access to the stream, allowing to search for a
*    specific item. Returns the position in the stream if the item is
*    found.
*************************************************************************/

bigsortcount_t bigsortfind(bigsorthandle_t *bsh, const void *cont, unsigned int lcont,
                           int (*compare)(void *uptr,
                                          const void *a, unsigned int la,
                                          const void *b, unsigned int lb),
                           void *uPtr, void *dest)

/************************************************************************/
{
    bigsortcount_t lower, upper;
    void *temp;
    unsigned int len, btype;

    temp = NULL;
    len = bsh->BSH_Len;

    lower = 0;
    upper = bsh->BSH_Count;

    btype = bsh->BSH_Type & BIGSORT_TYPE_TMASK;

    if(btype == BIGSORT_TYPE_TFILELIST) {
	if(bsh->BSH_Type & (BIGSORT_TYPE_FVARLENGTH | BIGSORT_TYPE_FVARLENGTH2)) {
            JSSETERROR(_ERR_BIGSORTGET, 1);
            EGOTO(error);
        }

        if((temp =
#ifdef CONFIG_USE_ALLOCA
            alloca(len)
#else
            malloc(len)
#endif
            ) == NULL)
            EGOTO(error);

        bsh->BSH_Type &= ~BIGSORT_TYPE_FFILEPOSOK; /*  be on the safe side  */

    } else if(btype != BIGSORT_TYPE_TMEMLIST) {
	EGOTO(error);
    }

    while(lower < upper) {
	const bigsortdata_t *p;
	bigsortcount_t m;
	unsigned int lp;
	int cmp;

	m = (lower + upper) / 2;

	if(btype == BIGSORT_TYPE_TMEMLIST) {
	    /* **  Binary search in memory list  ** */
	    p  = bsh->BSH_Buff[m].BSIE_Data;
	    lp = bsh->BSH_Buff[m].BSIE_Length;

	} else {
	    /* **  Binary search in file  ** */
            if(fseek(bsh->BSH_File, (off_t)m * len, SEEK_SET) == -1 ||
               fread(temp, len, 1, bsh->BSH_File) != 1)
                EGOTO(error);

            bsh->BSH_TotalRead += len;

	    p = temp;
	    lp = len;
	}

	cmp = compare(uPtr, p, lp, cont, lcont);
	if(bsh->BSH_Type & BIGSORT_TYPE_FINVERTED)
	    cmp = -cmp;

	if(cmp < 0)
	    lower = m + 1;
	else
	    upper = m;
    }

    if(dest != NULL && lower < bsh->BSH_Count) {
	if(btype == BIGSORT_TYPE_TMEMLIST) {
	    memcpy(dest, bsh->BSH_Buff[lower].BSIE_Data, bsh->BSH_Buff[lower].BSIE_Length);

	} else {
            if(fseek(bsh->BSH_File, (off_t)lower * len, SEEK_SET) == -1 ||
               fread(dest, len, 1, bsh->BSH_File) != 1)
                EGOTO(error);

            bsh->BSH_TotalRead += len;
	}
    }

#ifndef CONFIG_USE_ALLOCA
    if(temp) { free(temp);  temp = NULL; }
#endif

    return lower;
error:
#ifndef CONFIG_USE_ALLOCA
    if(temp) free(temp);
#endif
    return -1;
}
