/*JS*********************************************************************
 *
 *    Program : TESTBIGSORT
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : Test bigsort routines
 *
 *              Copyright (C) 1994-2002 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*#define CONFIG_USELOCALE*/

#ifdef CONFIG_USELOCALE
# include <locale.h>
#endif

#include <jssubs.h>

#define DEBUG

/*********     DEFINES                                          *********/
#ifdef DEBUG
#define OPTSTRING        "hn:N:m:l:udgVvD:"
#else
#define OPTSTRING        "hn:N:m:l:udgVv"
#endif

#define DEFLINELENGTH    100

#define MODE_DOSORT      1
#define MODE_VERBOSE     2

/*********     PROTOTYPES                                       *********/
static char *Buffer;
static unsigned int BuffSize;
static unsigned int Len = DEFLINELENGTH;

static int doGet(void *uptr, void *buff, unsigned int lbuff, bigsortcount_t nr)
{
    unsigned int l;

    if(fgets(Buffer, BuffSize, (FILE *)uptr) == NULL) {
	if(!feof((FILE *)uptr))
	    return -1;

	return -2; /*  EOF  */
    }

    if((l = strlen(Buffer)) > 0 && Buffer[l - 1] == '\n')
	--l;

    if(l >= lbuff)
	l = lbuff - 1;

    memcpy(buff, Buffer, l);
    ((char *)buff)[l] = '\0';

    return lbuff;
}

static int doCompare(void *uptr, const void *a, unsigned int la,
		     const void *b, unsigned int lb)
{
#ifdef CONFIG_USELOCALE
    return strcoll(a, b);
#else
    return strcmp(a, b);
#endif
}

int main(int argc, char *argv[])
{
    bigsorthandle_t bsh;
    FILE *fpIn, *fpOut;
    char *string;
    int c, mode, bMode, rlen;
    bigsortcount_t nExpected;

#ifdef CONFIG_USELOCALE
    setlocale(LC_ALL, "");
#endif

    BSortTempDir = getenv("TMPDIR");

    mode = MODE_DOSORT;
    bMode = 0;
    nExpected = -1;
    while((c = jsGetOpt(&argc, (void *)argv, OPTSTRING, (const char **)&string)) != EOF)
	switch(c) {
	case 'h':
	    goto help;

	case 'n':
	    BSortNSeqs = atoi(string);
	    break;

	case 'N':
	    nExpected = atoi(string);
	    break;

	case 'm':
	    BSortMaxMem = strtoul(string, NULL, 10);
	    break;

	case 'l':
	    Len = atoi(string);
	    break;

	case 'u':
	    mode &= ~MODE_DOSORT;
	    break;

	case 'd':
	    if(bMode & BIGSORT_FDISTINCT)
		bMode = (bMode & ~BIGSORT_FDISTINCT) | BIGSORT_FDISTINCT2;
	    else
		bMode = (bMode & ~BIGSORT_FDISTINCT2) | BIGSORT_FDISTINCT;
	    break;

	case 'g':
	    bMode |= BIGSORT_FGROUP;
	    break;

	case 'V':
	    bMode |= BIGSORT_FVARLEN;
	    break;

	case 'v':
	    mode |= MODE_VERBOSE;
	    break;

#ifdef DEBUG
	case 'D':
	    Debug = strtol(string, NULL, 0);
	    break;
#endif
	}

    if(argc != 2 && argc != 3)
	goto help;

    if(strcmp(argv[1], "-") == 0)
	fpIn = stdin;
    else if((fpIn = fopen(argv[1], "r")) == NULL)
	EGOTO(error);

    if(argc == 2 || strcmp(argv[2], "-") == 0)
	fpOut = stdout;
    else if((fpOut = fopen(argv[2], "w")) == NULL)
	EGOTO(error);

    /*  Now sort and output the sorted lines  */
    BuffSize = MAX(4096, Len + 1);
    if((Buffer = malloc(BuffSize)) == NULL)
	EGOTO(error);

    if(bigsortopen(&bsh, Len, nExpected, doGet,
		   (mode & MODE_DOSORT) ? doCompare : NULL, fpIn, bMode) == NULL)
	EGOTO(error);

    while((rlen = bigsortget(&bsh, -1, Buffer, Len)) >= 0) {
	if(mode & MODE_VERBOSE)
	    fprintf(fpOut, "%ld '%s'\n", bigsortcurrpos(&bsh) - 1, Buffer);
	else
	    fprintf(fpOut, "%s\n", Buffer);
    }

    free(Buffer);

    fprintf(stderr,
	    "Stats: %ldMB/%ldMB written/read, %ldMB max temp file size, %.0f average run length, %d temp files, %d iterations.\n",
	    (long)(bsh.BSH_TotalWritten >> 20), (long)(bsh.BSH_TotalRead >> 20),
	    (long)(bsh.BSH_MaxTempSize >> 20),
	    bsh.BSH_TotalRuns == 0 ? (double)bsh.BSH_Count :
	    (double)bsh.BSH_Count / (double)bsh.BSH_TotalRuns,
	    bsh.BSH_NTempFiles, bsh.BSH_MaxLevel);

    /*  Clean up  */
    if(bigsortclose(&bsh) == -1)
	EGOTO(error);

    if(fpOut != stdout)
	fclose(fpOut);

    if(fpIn != stdin)
	fclose(fpIn);

    return 0;

 error:
    jsperror("testbigsort");
    return 30;

 help:
    fprintf(stderr, "Usage: testbigsort [<opt>] <input> [<output>]\n\
  Sort lines from the input file to the output file or standard output\n\
  \"-\" reads standard input. The method is multi-phase sorting on temporary\n\
  files which allows the sorting of huge amounts of data that do not fit\n\
  into main memory.\n\
  Available options are:\n\
    -h         Show this help information.\n\
    -l<maxlen> Set the maximum line length. (Default is %d).\n\
    -n<nseqs>  Set the number of temporary files that are used in sorting\n\
               (Default is %d).\n\
    -N<nlines> Set the number of expected lines.\n\
    -u         Do not sort.\n\
    -d         Remove duplicate lines. If given twice, use the old method of\n\
               removing duplicate lines (which allows also direct access).\n\
    -g         Do grouping.\n\
    -V         Use variable length sorting, '-l' only specifies the maximum\n\
               length.\n\
    -v         Be verbose on output: Print also the return value of\n\
               'bigsortget'.\n\
    -m<maxmem> Set the maximum used memory for the initial in-memory sorting.\n\
               Defaults to %ld Bytes.\n", Len, BSortNSeqs, BSortMaxMem);
    return 5;
}
